# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
"""Implementation of a general "block vector"


The `pyomo.contrib.pynumero.sparse.block_vector` module includes methods
that extend linear algebra operations in numpy for case of structured
problems where linear algebra operations present an inherent block
structure.  This interface consider vectors of the form:

.. math::

   v = [v_1, v_2, v_3, ... , v_n]

where `v_i` are numpy arrays of dimension 1

.. rubric:: Contents

Methods specific to :py:class:`BlockVector`:

   * :py:meth:`~BlockVector.set_block`
   * :py:meth:`~BlockVector.get_block`
   * :py:meth:`~BlockVector.block_sizes`
   * :py:meth:`~BlockVector.get_block_size`
   * :py:meth:`~BlockVector.is_block_defined`
   * :py:meth:`~BlockVector.copyfrom`
   * :py:meth:`~BlockVector.copyto`
   * :py:meth:`~BlockVector.copy_structure`
   * :py:meth:`~BlockVector.set_blocks`
   * :py:meth:`~BlockVector.pprint`

Attributes specific to :py:class:`BlockVector`:

   * :py:attr:`~BlockVector.nblocks`
   * :py:attr:`~BlockVector.bshape`
   * :py:attr:`~BlockVector.has_none`


NumPy compatible methods:

   * :py:meth:`~numpy.ndarray.dot`
   * :py:meth:`~numpy.ndarray.sum`
   * :py:meth:`~numpy.ndarray.all`
   * :py:meth:`~numpy.ndarray.any`
   * :py:meth:`~numpy.ndarray.max`
   * :py:meth:`~numpy.ndarray.astype`
   * :py:meth:`~numpy.ndarray.clip`
   * :py:meth:`~numpy.ndarray.compress`
   * :py:meth:`~numpy.ndarray.conj`
   * :py:meth:`~numpy.ndarray.conjugate`
   * :py:meth:`~numpy.ndarray.nonzero`
   * :py:meth:`~numpy.ndarray.ptp` (NumPy 1.x only)
   * :py:meth:`~numpy.ndarray.round`
   * :py:meth:`~numpy.ndarray.std`
   * :py:meth:`~numpy.ndarray.var`
   * :py:meth:`~numpy.ndarray.tofile`
   * :py:meth:`~numpy.ndarray.min`
   * :py:meth:`~numpy.ndarray.mean`
   * :py:meth:`~numpy.ndarray.prod`
   * :py:meth:`~numpy.ndarray.fill`
   * :py:meth:`~numpy.ndarray.tolist`
   * :py:meth:`~numpy.ndarray.flatten`
   * :py:meth:`~numpy.ndarray.ravel`
   * :py:meth:`~numpy.ndarray.argmax`
   * :py:meth:`~numpy.ndarray.argmin`
   * :py:meth:`~numpy.ndarray.cumprod`
   * :py:meth:`~numpy.ndarray.cumsum`
   * :py:meth:`~numpy.ndarray.copy`

For example,

.. code-block:: python

   >>> import numpy as np
   >>> from pyomo.contrib.pynumero.sparse import BlockVector
   >>> v = BlockVector(2)
   >>> v.set_block(0, np.random.normal(size=100))
   >>> v.set_block(1, np.random.normal(size=30))
   >>> avg = v.mean()

NumPy compatible functions:

   * :py:func:`~numpy.log10`
   * :py:func:`~numpy.sin`
   * :py:func:`~numpy.cos`
   * :py:func:`~numpy.exp`
   * :py:func:`~numpy.ceil`
   * :py:func:`~numpy.floor`
   * :py:func:`~numpy.tan`
   * :py:func:`~numpy.arctan`
   * :py:func:`~numpy.arcsin`
   * :py:func:`~numpy.arccos`
   * :py:func:`~numpy.sinh`
   * :py:func:`~numpy.cosh`
   * :py:func:`~numpy.abs`
   * :py:func:`~numpy.tanh`
   * :py:func:`~numpy.arccosh`
   * :py:func:`~numpy.arcsinh`
   * :py:func:`~numpy.arctanh`
   * :py:func:`~numpy.fabs`
   * :py:func:`~numpy.sqrt`
   * :py:func:`~numpy.log`
   * :py:func:`~numpy.log2`
   * :py:func:`~numpy.absolute`
   * :py:func:`~numpy.isfinite`
   * :py:func:`~numpy.isinf`
   * :py:func:`~numpy.isnan`
   * :py:func:`~numpy.log1p`
   * :py:func:`~numpy.logical_not`
   * :py:func:`~numpy.expm1`
   * :py:func:`~numpy.exp2`
   * :py:func:`~numpy.sign`
   * :py:func:`~numpy.rint`
   * :py:func:`~numpy.square`
   * :py:func:`~numpy.positive`
   * :py:func:`~numpy.negative`
   * :py:func:`~numpy.rad2deg`
   * :py:func:`~numpy.deg2rad`
   * :py:func:`~numpy.conjugate`
   * :py:func:`~numpy.reciprocal`
   * :py:func:`~numpy.signbit`
   * :py:func:`~numpy.add`
   * :py:func:`~numpy.multiply`
   * :py:func:`~numpy.divide`
   * :py:func:`~numpy.subtract`
   * :py:func:`~numpy.greater`
   * :py:func:`~numpy.greater_equal`
   * :py:func:`~numpy.less`
   * :py:func:`~numpy.less_equal`
   * :py:func:`~numpy.not_equal`
   * :py:func:`~numpy.maximum`
   * :py:func:`~numpy.minimum`
   * :py:func:`~numpy.fmax`
   * :py:func:`~numpy.fmin`
   * :py:func:`~numpy.equal`
   * :py:func:`~numpy.logical_and`
   * :py:func:`~numpy.logical_or`
   * :py:func:`~numpy.logical_xor`
   * :py:func:`~numpy.logaddexp`
   * :py:func:`~numpy.logaddexp2`
   * :py:func:`~numpy.remainder`
   * :py:func:`~numpy.heaviside`
   * :py:func:`~numpy.hypot`

For example,

.. code-block:: python

   >>> import numpy as np
   >>> from pyomo.contrib.pynumero.sparse import BlockVector
   >>> v = BlockVector(2)
   >>> v.set_block(0, np.random.normal(size=100))
   >>> v.set_block(1, np.random.normal(size=30))
   >>> inf_norm = np.max(np.abs(v))

.. autosummary::

   BlockVector
   BlockVector.set_block
   BlockVector.get_block
   BlockVector.block_sizes
   BlockVector.get_block_size
   BlockVector.is_block_defined
   BlockVector.copyfrom
   BlockVector.copyto
   BlockVector.copy_structure
   BlockVector.set_blocks
   BlockVector.pprint
   BlockVector.nblocks
   BlockVector.bshape
   BlockVector.has_none

"""

import operator

from ..dependencies import numpy as np
from .base_block import (
    BaseBlockVector,
    vec_unary_ufuncs,
    vec_binary_ufuncs,
    vec_associative_reductions,
)


class NotFullyDefinedBlockVectorError(Exception):
    pass


def assert_block_structure(vec):
    if vec.has_none:
        msg = 'Operation not allowed with None blocks.'
        raise NotFullyDefinedBlockVectorError(msg)


class BlockVector(BaseBlockVector, np.ndarray):
    """
    Structured vector interface. This interface can be used to
    perform operations on vectors composed by vectors. For example,

    >>> import numpy as np
    >>> from pyomo.contrib.pynumero.sparse import BlockVector
    >>> bv = BlockVector(3)
    >>> v0 = np.ones(3)
    >>> v1 = v0*2
    >>> v2 = np.random.normal(size=4)
    >>> bv.set_block(0, v0)
    >>> bv.set_block(1, v1)
    >>> bv.set_block(2, v2)
    >>> bv2 = BlockVector(2)
    >>> bv2.set_block(0, v0)
    >>> bv2.set_block(1, bv)

    Attributes
    ----------
    _nblocks: int
        number of blocks
    _brow_lengths: numpy.ndarray
        1D-Array of size nblocks that specifies the length of each entry
        in the block vector
    _undefined_brows: set
        A set of block indices for which the blocks are still None (i.e., the dimensions
        have not yet ben set). Operations with BlockVectors require all entries to be
        different than None.

    Parameters
    ----------
    nblocks: int
        The number of blocks in the BlockVector

    """

    def __new__(cls, nblocks):
        blocks = [None for i in range(nblocks)]
        arr = np.asarray(blocks, dtype='object')
        obj = arr.view(cls)
        obj._nblocks = nblocks

        obj._brow_lengths = np.empty(nblocks, dtype=np.float64)
        obj._brow_lengths.fill(np.nan)
        obj._undefined_brows = set(range(nblocks))

        return obj

    def __init__(self, nblocks):
        pass

    def __array_finalize__(self, obj):
        """This method is required to subclass from numpy array"""
        if obj is None:
            return
        self._brow_lengths = getattr(obj, '_brow_lengths', None)
        self._nblocks = getattr(obj, '_nblocks', 0)
        self._undefined_brows = getattr(obj, '_undefined_brows', None)

    def __array_prepare__(self, out_arr, context=None):
        """This method is required to subclass from numpy array"""
        return super(BlockVector, self).__array_prepare__(self, out_arr, context)

    def __array_wrap__(self, out_arr, context=None):
        """This method is required to subclass from numpy array"""
        return super(BlockVector, self).__array_wrap__(self, out_arr, context)

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        """Runs ufuncs specializations to BlockVector"""
        if kwargs.get('out', None) is not None:
            return NotImplemented
        if method == 'reduce' and ufunc in vec_associative_reductions:
            (arg,) = inputs
            return self._reduction_operation(ufunc, method, arg, kwargs)
        if method == '__call__':
            if ufunc in vec_unary_ufuncs:
                (arg,) = inputs
                return self._unary_operation(ufunc, method, arg, kwargs)
            if ufunc in vec_binary_ufuncs:
                return self._binary_operation(ufunc, method, inputs, kwargs)
        return NotImplemented

    def _reduction_operation(self, ufunc, method, x, kwargs):
        results = [
            self._unary_operation(ufunc, method, x.get_block(i), kwargs)
            for i in range(x.nblocks)
        ]
        if len(results) == 1:
            return results[0]
        else:
            return super().__array_ufunc__(ufunc, method, np.array(results), **kwargs)

    def _unary_operation(self, ufunc, method, x, kwargs):
        """Run recursion to perform unary_funcs on BlockVector"""
        # ToDo: deal with out
        if isinstance(x, BlockVector):
            v = BlockVector(x.nblocks)
            for i in range(x.nblocks):
                v.set_block(
                    i, self._unary_operation(ufunc, method, x.get_block(i), kwargs)
                )
            return v
        elif type(x) == np.ndarray:
            return super().__array_ufunc__(ufunc, method, x, **kwargs)
        else:
            return NotImplemented

    def _binary_operation(self, ufunc, method, args, kwargs):
        """Run recursion to perform binary_funcs on BlockVector"""
        # ToDo: deal with out
        x1, x2 = args
        if isinstance(x1, BlockVector) and isinstance(x2, BlockVector):
            assert_block_structure(x1)
            assert_block_structure(x2)
            assert (
                x1.nblocks == x2.nblocks
            ), 'Operation on BlockVectors need the same number of blocks on each operand'
            assert x1.size == x2.size, 'Dimension mismatch {}!={}'.format(
                x1.size, x2.size
            )
            res = BlockVector(x1.nblocks)

            for i in range(x1.nblocks):
                _args = (x1.get_block(i), x2.get_block(i))
                res.set_block(i, self._binary_operation(ufunc, method, _args, kwargs))
            return res
        elif type(x1) == np.ndarray and isinstance(x2, BlockVector):
            assert_block_structure(x2)
            assert x1.size == x2.size, 'Dimension mismatch {}!={}'.format(
                x1.size, x2.size
            )
            res = BlockVector(x2.nblocks)
            accum = 0
            for i in range(x2.nblocks):
                nelements = x2._brow_lengths[i]
                _args = (x1[accum : accum + nelements], x2.get_block(i))
                res.set_block(i, self._binary_operation(ufunc, method, _args, kwargs))
                accum += nelements
            return res
        elif type(x2) == np.ndarray and isinstance(x1, BlockVector):
            assert_block_structure(x1)
            assert x1.size == x2.size, 'Dimension mismatch {}!={}'.format(
                x1.size, x2.size
            )
            res = BlockVector(x1.nblocks)
            accum = 0
            for i in range(x1.nblocks):
                nelements = x1._brow_lengths[i]
                _args = (x1.get_block(i), x2[accum : accum + nelements])
                res.set_block(i, self._binary_operation(ufunc, method, _args, kwargs))
                accum += nelements
            return res
        elif np.isscalar(x1) and isinstance(x2, BlockVector):
            assert_block_structure(x2)
            res = BlockVector(x2.nblocks)
            for i in range(x2.nblocks):
                _args = (x1, x2.get_block(i))
                res.set_block(i, self._binary_operation(ufunc, method, _args, kwargs))
            return res
        elif np.isscalar(x2) and isinstance(x1, BlockVector):
            assert_block_structure(x1)
            res = BlockVector(x1.nblocks)
            for i in range(x1.nblocks):
                _args = (x1.get_block(i), x2)
                res.set_block(i, self._binary_operation(ufunc, method, _args, kwargs))
            return res
        elif (type(x1) == np.ndarray or np.isscalar(x1)) and (
            type(x2) == np.ndarray or np.isscalar(x2)
        ):
            return super(BlockVector, self).__array_ufunc__(
                ufunc, method, *args, **kwargs
            )
        else:
            if x1.__class__.__name__ == 'MPIBlockVector':
                raise RuntimeError('Operation not supported by BlockVector')
            if x2.__class__.__name__ == 'MPIBlockVector':
                raise RuntimeError('Operation not supported by BlockVector')
            return NotImplemented

    @property
    def nblocks(self):
        """
        Returns the number of blocks.
        """
        return self._nblocks

    @property
    def bshape(self):
        """
        Returns the number of blocks in this BlockVector in a tuple.
        """
        return (self.nblocks,)

    @property
    def shape(self):
        """
        Returns total number of elements in this BlockVector
        """
        assert_block_structure(self)
        return (np.sum(self._brow_lengths),)

    @property
    def size(self):
        """
        Returns total number of elements in this BlockVector
        """
        assert_block_structure(self)
        return np.sum(self._brow_lengths)

    @property
    def ndim(self):
        """
        Returns dimension of this BlockVector
        """
        return 1

    @property
    def has_none(self):
        """
        Indicate if this BlockVector has any none entries.
        """
        # this flag is updated in __setattr__
        return len(self._undefined_brows) != 0

    def block_sizes(self, copy=True):
        """
        Returns 1D-Array with sizes of individual blocks in this BlockVector
        """
        assert_block_structure(self)
        if copy:
            return self._brow_lengths.copy()
        return self._brow_lengths

    def get_block_size(self, ndx):
        if ndx in self._undefined_brows:
            raise NotFullyDefinedBlockVectorError(
                'The dimensions of the requested block are not defined.'
            )
        return int(self._brow_lengths[ndx])

    def _set_block_size(self, ndx, size):
        if ndx in self._undefined_brows:
            self._undefined_brows.remove(ndx)
            self._brow_lengths[ndx] = size
            if len(self._undefined_brows) == 0:
                self._brow_lengths = np.asarray(self._brow_lengths, dtype=np.int64)
        else:
            if self._brow_lengths[ndx] != size:
                raise ValueError(
                    'Incompatible dimensions for '
                    'block {ndx}; got {got}; '
                    'expected {exp}'.format(
                        ndx=ndx, got=size, exp=self._brow_lengths[ndx]
                    )
                )

    def is_block_defined(self, ndx):
        return ndx not in self._undefined_brows

    def dot(self, other, out=None):
        """
        Returns dot product

        Parameters
        ----------
        other : ndarray or BlockVector

        Returns
        -------
        float

        """
        assert out is None, 'Operation not supported with out keyword'
        assert_block_structure(self)
        if isinstance(other, BlockVector):
            assert_block_structure(other)
            assert self.shape == other.shape, 'Dimension mismatch {} != {}'.format(
                self.shape, other.shape
            )
            assert (
                self.nblocks == other.nblocks
            ), 'Number of blocks mismatch {} != {}'.format(self.nblocks, other.nblocks)
            return sum(
                self.get_block(i).dot(other.get_block(i)) for i in range(self.nblocks)
            )
        elif type(other) == np.ndarray:
            bv = self.flatten()
            return bv.dot(other)
        else:
            if other.__class__.__name__ == 'MPIBlockVector':
                raise RuntimeError('Operation not supported by BlockVector')
            raise NotImplementedError()

    def sum(self, axis=None, dtype=None, out=None, keepdims=False):
        """
        Returns the sum of all entries in this BlockVector
        """
        assert_block_structure(self)
        results = np.array([self.get_block(i).sum() for i in range(self.nblocks)])
        return results.sum(axis=axis, dtype=dtype, out=out, keepdims=keepdims)

    def all(self, axis=None, out=None, keepdims=False):
        """
        Returns True if all elements evaluate to True.
        """
        assert_block_structure(self)
        results = np.array(
            [self.get_block(i).all() for i in range(self.nblocks)], dtype=bool
        )
        return results.all(axis=axis, out=out, keepdims=keepdims)

    def any(self, axis=None, out=None, keepdims=False):
        """
        Returns True if any element evaluate to True.
        """
        assert_block_structure(self)
        results = np.array(
            [self.get_block(i).any() for i in range(self.nblocks)], dtype=bool
        )
        return results.any(axis=axis, out=out, keepdims=keepdims)

    def max(self, axis=None, out=None, keepdims=False):
        """
        Returns the largest value stored in this BlockVector
        """
        assert_block_structure(self)
        results = list()
        for block in self:
            if block.size > 0:
                results.append(block.max())
        return max(results)

    def astype(self, dtype, order='K', casting='unsafe', subok=True, copy=True):
        """Copy of the array, cast to a specified type"""
        if copy:
            bv = BlockVector(self.nblocks)
            for bid, vv in enumerate(self):
                if bid not in self._undefined_brows:
                    bv.set_block(
                        bid,
                        vv.astype(
                            dtype, order=order, casting=casting, subok=subok, copy=copy
                        ),
                    )
            return bv
        raise NotImplementedError("astype not implemented for copy=False")

    def clip(self, min=None, max=None, out=None):
        """
        Return BlockVector whose values are limited to [min, max].
        One of max or min must be given.

        Parameters
        ----------
        min: scalar_like, optional
            Minimum value. If None, clipping is not performed on lower interval edge.
        max: scalar_like, optional
            Maximum value. If None, clipping is not performed on upper interval edge.

        Returns
        -------
        BlockVector

        """
        assert_block_structure(self)
        assert out is None, 'Out keyword not supported'

        bv = BlockVector(self.nblocks)
        for bid in range(self.nblocks):
            bv.set_block(bid, self.get_block(bid).clip(min=min, max=max, out=None))
        return bv

    def compress(self, condition, axis=None, out=None):
        """
        Return selected slices of each subblock.

        Parameters
        ----------
        condition: Array or BlockVector that selects which entries to return.
            Determines to select (evaluate True in condition)

        Returns
        -------
        BlockVector

        """
        assert_block_structure(self)
        assert out is None, 'Out keyword not supported'
        result = BlockVector(self.nblocks)

        if isinstance(condition, BlockVector):
            assert_block_structure(condition)
            assert self.shape == condition.shape, 'Dimension mismatch {} != {}'.format(
                self.shape, condition.shape
            )
            assert (
                self.nblocks == condition.nblocks
            ), 'Number of blocks mismatch {} != {}'.format(
                self.nblocks, condition.nblocks
            )
            for idx in range(self.nblocks):
                result.set_block(
                    idx, self.get_block(idx).compress(condition.get_block(idx))
                )
            return result
        elif type(condition) == np.ndarray:
            assert self.shape == condition.shape, 'Dimension mismatch {} != {}'.format(
                self.shape, condition.shape
            )
            accum = 0
            for idx in range(self.nblocks):
                nelements = self._brow_lengths[idx]
                result.set_block(
                    idx,
                    self.get_block(idx).compress(condition[accum : accum + nelements]),
                )
                accum += nelements
            return result
        else:
            if other.__class__.__name__ == 'MPIBlockVector':
                raise RuntimeError('Operation not supported by BlockVector')
            raise NotImplementedError()

    def conj(self):
        """
        Complex-conjugate all elements.
        """
        assert_block_structure(self)
        result = BlockVector(self.nblocks)
        for idx in range(self.nblocks):
            result.set_block(idx, self.get_block(idx).conj())
        return result

    def conjugate(self):
        """
        Complex-conjugate all elements.
        """
        assert_block_structure(self)
        result = BlockVector(self.nblocks)
        for idx in range(self.nblocks):
            result.set_block(idx, self.get_block(idx).conjugate())
        return result

    def nonzero(self):
        """
        Return the indices of the elements that are non-zero.
        """
        assert_block_structure(self)
        result = BlockVector(self.nblocks)
        for idx in range(self.nblocks):
            result.set_block(idx, self.get_block(idx).nonzero()[0])
        return (result,)

    if np.__version__[0] < '2':

        def ptp(self, axis=None, out=None, keepdims=False):
            """
            Peak to peak (maximum - minimum) value along a given axis.
            """
            assert_block_structure(self)
            assert out is None, 'Out keyword not supported'
            return self.max() - self.min()

    def round(self, decimals=0, out=None):
        """
        Return BlockVector with each element rounded to the given number of decimals
        """
        assert_block_structure(self)
        assert out is None, 'Out keyword not supported'
        result = BlockVector(self.nblocks)
        for idx in range(self.nblocks):
            result.set_block(idx, self.get_block(idx).round(decimals=decimals))
        return result

    def std(self, axis=None, dtype=None, out=None, ddof=0, keepdims=False):
        """
        Returns the standard deviation of the BlockVector elements.
        """
        return self.flatten().std(
            axis=axis, dtype=dtype, out=out, ddof=ddof, keepdims=keepdims
        )

    def var(self, axis=None, dtype=None, out=None, ddof=0, keepdims=False):
        """
        Returns the variance of the BlockVector elements.
        """
        return self.flatten().var(
            axis=axis, dtype=dtype, out=out, ddof=ddof, keepdims=keepdims
        )

    def tofile(self, fid, sep="", format="%s"):
        """
        Writes flat version of BlockVector to a file as text or binary (default).
        """
        self.flatten().tofile(fid, sep=sep, format=format)

    def min(self, axis=None, out=None, keepdims=False):
        """
        Returns the smallest value stored in the vector
        """
        assert_block_structure(self)
        results = list()
        for block in self:
            if block.size > 0:
                results.append(block.min())
        return min(results)

    def mean(self, axis=None, dtype=None, out=None, keepdims=False):
        """
        Returns the average of all entries in this BlockVector
        """
        n = self.size
        if n == 0:
            return 0.0
        return self.sum(axis=axis, dtype=dtype, out=out, keepdims=keepdims) / n

    def prod(self, axis=None, dtype=None, out=None, keepdims=False):
        """
        Returns the product of all entries in this BlockVector
        """
        assert_block_structure(self)
        results = np.array([self.get_block(i).prod() for i in range(self.nblocks)])
        return results.prod(axis=axis, dtype=dtype, out=out, keepdims=keepdims)

    def fill(self, value):
        """
        Fills the BlockVector with a scalar value.

        Parameters
        ----------
        value : scalar
            All elements in the vector will be assigned this value

        Returns
        -------
        None

        """
        assert_block_structure(self)
        for i in range(self.nblocks):
            self.get_block(i).fill(value)

    def tolist(self):
        """
        Return the BlockVector flattened as a list.

        Returns
        -------
        list

        """
        return self.flatten().tolist()

    def flatten(self, order='C'):
        """
        Converts the BlockVector to a NumPy array. This will also call flatten on the underlying NumPy arrays in
        the BlockVector.

        Parameters
        ----------
        order: str: {C, F, A, K}, optional
            See NumPy array documentation.

        Returns
        -------
        flat_array: numpy.ndarray
            The NumPy array resulting from concatenating all of the blocks
        """
        assert_block_structure(self)
        all_blocks = tuple(
            self.get_block(i).flatten(order=order) for i in range(self.nblocks)
        )
        return np.concatenate(all_blocks)

    def ravel(self, order='C'):
        """
        Converts the BlockVector into a NumPy array. Note that ravel is also called
        on all of the NumPy arrays in the BlockVector before concatenating them.

        Parameters
        ----------
        order: str
            See NumPy documentation.

        Returns
        -------
        res: numpy.ndarray
        """
        assert_block_structure(self)
        all_blocks = tuple(
            self.get_block(i).ravel(order=order) for i in range(self.nblocks)
        )
        return np.concatenate(all_blocks)

    def argmax(self, axis=None, out=None):
        """
        Returns the index of the largest element.
        """
        assert_block_structure(self)
        return self.flatten().argmax(axis=axis, out=out)

    def argmin(self, axis=None, out=None):
        """
        Returns the index of the smallest element.
        """
        assert_block_structure(self)
        return self.flatten().argmin(axis=axis, out=out)

    def cumprod(self, axis=None, dtype=None, out=None):
        """
        Returns the cumulative product of the elements along the given axis.
        """
        flat = self.flatten().cumprod(axis=axis, dtype=dtype, out=out)
        v = self.clone()
        v.copyfrom(flat)
        return v

    def cumsum(self, axis=None, dtype=None, out=None):
        """
        Returns the cumulative sum of the elements along the given axis.
        """
        flat = self.flatten().cumsum(axis=axis, dtype=dtype, out=out)
        v = self.clone()
        v.copyfrom(flat)
        return v

    def clone(self, value=None, copy=True):
        """
        Returns a copy of this BlockVector

        Parameters
        ----------
        value: scalar (optional)
            all entries of the cloned vector are set to this value
        copy: bool (optional)
            if True makes a deepcopy of each block in this vector. default True

        Returns
        -------
        BlockVector

        """
        result = BlockVector(self.nblocks)
        for idx in range(self.nblocks):
            if idx not in self._undefined_brows:
                if copy:
                    result.set_block(idx, self.get_block(idx).copy())
                else:
                    result.set_block(idx, self.get_block(idx))
        if value is not None:
            result.fill(value)
        return result

    def copyfrom(self, other):
        """
        Copy entries of other vector into this vector

        Parameters
        ----------
        other: BlockVector or numpy.ndarray
            vector to be copied to this BlockVector

        Returns
        -------
        None

        """
        assert_block_structure(self)

        if isinstance(other, BlockVector):
            assert_block_structure(other)

            assert self.shape == other.shape, 'Dimension mismatch {} != {}'.format(
                self.shape, other.shape
            )
            assert (
                self.nblocks == other.nblocks
            ), 'Number of blocks mismatch {} != {}'.format(self.nblocks, other.nblocks)
            for idx in range(other.nblocks):
                if isinstance(self.get_block(idx), BlockVector):
                    self.get_block(idx).copyfrom(other.get_block(idx))
                elif isinstance(self.get_block(idx), np.ndarray):
                    if isinstance(other.get_block(idx), BlockVector):
                        self.set_block(idx, other.get_block(idx).copy())
                    elif isinstance(other.get_block(idx), np.ndarray):
                        np.copyto(self.get_block(idx), other.get_block(idx))
                    else:
                        raise RuntimeError('Input not recognized')
                elif self.get_block(idx) is None:
                    if isinstance(other.get_block(idx), np.ndarray):
                        # this include block vectors too
                        self.set_block(idx, other.get_block(idx).copy())
                    else:
                        raise RuntimeError('Input not recognized')
                else:
                    raise RuntimeError('Input not recognized')
        elif isinstance(other, np.ndarray):
            assert self.shape == other.shape, 'Dimension mismatch {} != {}'.format(
                self.shape, other.shape
            )

            offset = 0
            for idx in range(self.nblocks):
                subarray = other[offset : offset + self.get_block(idx).size]
                if isinstance(self.get_block(idx), BlockVector):
                    self.get_block(idx).copyfrom(subarray)
                else:
                    np.copyto(self.get_block(idx), subarray)
                offset += self.get_block(idx).size
        else:
            raise NotImplementedError('Operation not supported by BlockVector')

    def copyto(self, other):
        """
        Copy entries of this BlockVector into other

        Parameters
        ----------
        other: BlockVector or numpy.ndarray

        Returns
        -------
        None

        """

        if isinstance(other, BlockVector):
            msgj = 'Number of blocks mismatch {} != {}'.format(
                self.nblocks, other.nblocks
            )
            assert self.nblocks == other.nblocks, msgj
            for idx in range(self.nblocks):
                if isinstance(other.get_block(idx), BlockVector):
                    other.get_block(idx).copyfrom(self.get_block(idx))
                elif isinstance(other.get_block(idx), np.ndarray):
                    if self.get_block(idx) is not None:
                        np.copyto(other.get_block(idx), self.get_block(idx).flatten())
                    else:
                        other.set_block(idx, None)
                elif other.get_block(idx) is None:
                    if self.get_block(idx) is not None:
                        other.set_block(idx, self.get_block(idx).copy())
                    else:
                        other.set_block(idx, None)
                else:
                    raise RuntimeError('Should never get here')

        elif isinstance(other, np.ndarray):
            np.copyto(other, self.flatten())
        else:
            raise NotImplementedError()

    def copy(self, order='C'):
        """
        Returns a copy of the BlockVector
        """
        bv = BlockVector(self.nblocks)
        for bid in range(self.nblocks):
            if bid not in self._undefined_brows:
                bv.set_block(bid, self.get_block(bid).copy(order=order))
        return bv

    def copy_structure(self):
        """
        Returns a copy of the BlockVector structure filled with zeros
        """
        bv = BlockVector(self.nblocks)
        for bid in range(self.nblocks):
            if self.get_block(bid) is not None:
                if isinstance(self.get_block(bid), BlockVector):
                    bv.set_block(bid, self.get_block(bid).copy_structure())
                elif type(self.get_block(bid)) == np.ndarray:
                    bv.set_block(
                        bid,
                        np.zeros(
                            self.get_block(bid).size, dtype=self.get_block(bid).dtype
                        ),
                    )
                else:
                    raise NotImplementedError('Should never get here')
        return bv

    def set_blocks(self, blocks):
        """
        Assigns vectors in blocks

        Parameters
        ----------
        blocks: list
            list of numpy.ndarrays and/or BlockVectors

        Returns
        -------
        None

        """
        assert isinstance(blocks, list), 'blocks should be passed in ordered list'
        assert (
            len(blocks) == self.nblocks
        ), 'More blocks passed than allocated {} != {}'.format(
            len(blocks), self.nblocks
        )
        for idx, blk in enumerate(blocks):
            self.set_block(idx, blk)

    def __iter__(self):
        for ndx in range(self._nblocks):
            yield self.get_block(ndx)

    def __add__(self, other):
        # add this BlockVector with other vector
        # supports addition with scalar, numpy.ndarray and BlockVectors
        # returns BlockVector
        result = BlockVector(self.nblocks)
        assert_block_structure(self)
        if isinstance(other, BlockVector):
            assert_block_structure(other)
            assert self.shape == other.shape, 'Dimension mismatch {} != {}'.format(
                self.shape, other.shape
            )
            assert (
                self.nblocks == other.nblocks
            ), 'Number of blocks mismatch {} != {}'.format(self.nblocks, other.nblocks)
            for idx, blk in enumerate(self):
                result.set_block(idx, blk + other.get_block(idx))
            return result
        elif type(other) == np.ndarray:
            assert self.shape == other.shape, 'Dimension mismatch {} != {}'.format(
                self.shape, other.shape
            )
            accum = 0
            for idx, blk in enumerate(self):
                nelements = self._brow_lengths[idx]
                result.set_block(idx, blk + other[accum : accum + nelements])
                accum += nelements
            return result
        elif np.isscalar(other):
            for idx, blk in enumerate(self):
                result.set_block(idx, blk + other)
            return result
        else:
            if other.__class__.__name__ == 'MPIBlockVector':
                raise RuntimeError('Operation not supported by BlockVector')
            raise NotImplementedError()

    def __radd__(self, other):  # other + self
        return self.__add__(other)

    def __sub__(self, other):
        # subtract this BlockVector with other vector
        # supports subtraction with scalar, numpy.ndarray and BlockVectors
        # returns BlockVector
        result = BlockVector(self.nblocks)
        assert_block_structure(self)
        if isinstance(other, BlockVector):
            assert_block_structure(other)
            assert self.shape == other.shape, 'Dimension mismatch {} != {}'.format(
                self.shape, other.shape
            )
            assert (
                self.nblocks == other.nblocks
            ), 'Number of blocks mismatch {} != {}'.format(self.nblocks, other.nblocks)
            for idx, blk in enumerate(self):
                result.set_block(idx, blk - other.get_block(idx))
            return result
        elif type(other) == np.ndarray:
            assert self.shape == other.shape, 'Dimension mismatch {} != {}'.format(
                self.shape, other.shape
            )
            accum = 0
            for idx, blk in enumerate(self):
                nelements = self._brow_lengths[idx]
                result.set_block(idx, blk - other[accum : accum + nelements])
                accum += nelements
            return result
        elif np.isscalar(other):
            for idx, blk in enumerate(self):
                result.set_block(idx, blk - other)
            return result
        else:
            if other.__class__.__name__ == 'MPIBlockVector':
                raise RuntimeError('Operation not supported by BlockVector')
            raise NotImplementedError()

    def __rsub__(self, other):  # other - self
        result = BlockVector(self.nblocks)
        assert_block_structure(self)
        if isinstance(other, BlockVector):
            assert_block_structure(other)
            assert self.shape == other.shape, 'Dimension mismatch {} != {}'.format(
                self.shape, other.shape
            )
            assert (
                self.nblocks == other.nblocks
            ), 'Number of blocks mismatch {} != {}'.format(self.nblocks, other.nblocks)
            for idx, blk in enumerate(self):
                result.set_block(idx, other.get_block(idx) - blk)
            return result

        elif type(other) == np.ndarray:
            assert self.shape == other.shape, 'Dimension mismatch {} != {}'.format(
                self.shape, other.shape
            )
            accum = 0
            for idx, blk in enumerate(self):
                nelements = self._brow_lengths[idx]
                result.set_block(idx, other[accum : accum + nelements] - blk)
                accum += nelements
            return result
        elif np.isscalar(other):
            for idx, blk in enumerate(self):
                result.set_block(idx, other - blk)
            return result
        else:
            if other.__class__.__name__ == 'MPIBlockVector':
                raise RuntimeError('Operation not supported by BlockVector')
            raise NotImplementedError()

    def __mul__(self, other):
        # elementwise multiply this BlockVector with other vector
        # supports multiplication with scalar, numpy.ndarray and BlockVectors
        # returns BlockVector
        assert_block_structure(self)
        result = BlockVector(self.nblocks)
        if isinstance(other, BlockVector):
            assert_block_structure(other)
            assert self.shape == other.shape, 'Dimension mismatch {} != {}'.format(
                self.shape, other.shape
            )
            assert (
                self.nblocks == other.nblocks
            ), 'Number of blocks mismatch {} != {}'.format(self.nblocks, other.nblocks)
            for idx, blk in enumerate(self):
                result.set_block(idx, blk.__mul__(other.get_block(idx)))
            return result
        elif type(other) == np.ndarray:
            assert self.shape == other.shape, 'Dimension mismatch {} != {}'.format(
                self.shape, other.shape
            )
            accum = 0
            for idx, blk in enumerate(self):
                nelements = self._brow_lengths[idx]
                result.set_block(idx, blk.__mul__(other[accum : accum + nelements]))
                accum += nelements
            return result
        elif np.isscalar(other):
            for idx, blk in enumerate(self):
                result.set_block(idx, blk.__mul__(other))
            return result
        else:
            if other.__class__.__name__ == 'MPIBlockVector':
                raise RuntimeError('Operation not supported by BlockVector')
            raise NotImplementedError()

    def __rmul__(self, other):  # other + self
        return self.__mul__(other)

    def __truediv__(self, other):
        # elementwise divide this BlockVector with other vector
        # supports division with scalar, numpy.ndarray and BlockVectors
        # returns BlockVector
        assert_block_structure(self)
        result = BlockVector(self.nblocks)
        if isinstance(other, BlockVector):
            assert_block_structure(other)
            assert self.shape == other.shape, 'Dimension mismatch {} != {}'.format(
                self.shape, other.shape
            )
            assert (
                self.nblocks == other.nblocks
            ), 'Number of blocks mismatch {} != {}'.format(self.nblocks, other.nblocks)
            for idx, blk in enumerate(self):
                result.set_block(idx, blk / other.get_block(idx))
            return result
        elif type(other) == np.ndarray:
            assert self.shape == other.shape, 'Dimension mismatch {} != {}'.format(
                self.shape, other.shape
            )
            accum = 0
            for idx, blk in enumerate(self):
                nelements = self._brow_lengths[idx]
                result.set_block(idx, blk / other[accum : accum + nelements])
                accum += nelements
            return result
        elif np.isscalar(other):
            for idx, blk in enumerate(self):
                result.set_block(idx, blk / other)
            return result
        else:
            if other.__class__.__name__ == 'MPIBlockVector':
                raise RuntimeError('Operation not supported by BlockVector')
            raise NotImplementedError()

    def __rtruediv__(self, other):
        assert_block_structure(self)
        result = BlockVector(self.nblocks)
        if isinstance(other, BlockVector):
            assert_block_structure(other)
            assert self.shape == other.shape, 'Dimension mismatch {} != {}'.format(
                self.shape, other.shape
            )
            assert (
                self.nblocks == other.nblocks
            ), 'Number of blocks mismatch {} != {}'.format(self.nblocks, other.nblocks)
            for idx, blk in enumerate(self):
                result.set_block(idx, other.get_block(idx) / blk)
            return result
        elif type(other) == np.ndarray:
            assert self.shape == other.shape, 'Dimension mismatch {} != {}'.format(
                self.shape, other.shape
            )
            accum = 0
            for idx, blk in enumerate(self):
                nelements = self._brow_lengths[idx]
                result.set_block(idx, other[accum : accum + nelements] / blk)
                accum += nelements
            return result
        elif np.isscalar(other):
            for idx, blk in enumerate(self):
                result.set_block(idx, other / blk)
            return result
        else:
            if other.__class__.__name__ == 'MPIBlockVector':
                raise RuntimeError('Operation not supported by BlockVector')
            raise NotImplementedError()

    def __floordiv__(self, other):
        assert_block_structure(self)
        result = BlockVector(self.nblocks)
        if isinstance(other, BlockVector):
            assert_block_structure(other)
            assert self.shape == other.shape, 'Dimension mismatch {} != {}'.format(
                self.shape, other.shape
            )
            assert (
                self.nblocks == other.nblocks
            ), 'Number of blocks mismatch {} != {}'.format(self.nblocks, other.nblocks)
            for idx, blk in enumerate(self):
                result.set_block(idx, blk // other.get_block(idx))
            return result
        elif type(other) == np.ndarray:
            assert self.shape == other.shape, 'Dimension mismatch {} != {}'.format(
                self.shape, other.shape
            )
            accum = 0
            for idx, blk in enumerate(self):
                nelements = self._brow_lengths[idx]
                result.set_block(idx, blk // other[accum : accum + nelements])
                accum += nelements
            return result
        elif np.isscalar(other):
            for idx, blk in enumerate(self):
                result.set_block(idx, blk // other)
            return result
        else:
            if other.__class__.__name__ == 'MPIBlockVector':
                raise RuntimeError('Operation not supported by BlockVector')
            raise NotImplementedError()

    def __rfloordiv__(self, other):
        assert_block_structure(self)
        result = BlockVector(self.nblocks)
        if isinstance(other, BlockVector):
            assert_block_structure(other)
            assert self.shape == other.shape, 'Dimension mismatch {} != {}'.format(
                self.shape, other.shape
            )
            assert (
                self.nblocks == other.nblocks
            ), 'Number of blocks mismatch {} != {}'.format(self.nblocks, other.nblocks)
            for idx, blk in enumerate(self):
                result.set_block(idx, other.get_block(idx) // blk)
            return result
        elif type(other) == np.ndarray:
            assert self.shape == other.shape, 'Dimension mismatch {} != {}'.format(
                self.shape, other.shape
            )
            accum = 0
            for idx, blk in enumerate(self):
                nelements = self._brow_lengths[idx]
                result.set_block(idx, other[accum : accum + nelements] // blk)
                accum += nelements
            return result
        elif np.isscalar(other):
            for idx, blk in enumerate(self):
                result.set_block(idx, other // blk)
            return result
        else:
            if other.__class__.__name__ == 'MPIBlockVector':
                raise RuntimeError('Operation not supported by BlockVector')
            raise NotImplementedError()

    def __iadd__(self, other):
        # elementwise inplace addition to this BlockVector with other vector
        # supports addition with scalar, numpy.ndarray and BlockVectors
        assert_block_structure(self)
        if np.isscalar(other):
            for idx, blk in enumerate(self):
                blk += other
            return self
        elif isinstance(other, BlockVector):
            assert_block_structure(other)
            assert self.shape == other.shape, 'Dimension mismatch {} != {}'.format(
                self.shape, other.shape
            )
            assert (
                self.nblocks == other.nblocks
            ), 'Number of blocks mismatch {} != {}'.format(self.nblocks, other.nblocks)
            for idx, blk in enumerate(self):
                blk += other.get_block(idx)
            return self
        elif type(other) == np.ndarray:
            assert self.shape == other.shape, 'Dimension mismatch {} != {}'.format(
                self.shape, other.shape
            )
            accum = 0
            for idx, blk in enumerate(self):
                nelements = self._brow_lengths[idx]
                blk += other[accum : accum + nelements]
                accum += nelements
            return self
        else:
            raise NotImplementedError()

    def __isub__(self, other):
        # elementwise inplace subtraction to this BlockVector with other vector
        # supports subtraction with scalar, numpy.ndarray and BlockVectors
        assert_block_structure(self)
        if np.isscalar(other):
            for idx, blk in enumerate(self):
                blk -= other
            return self
        elif isinstance(other, BlockVector):
            assert_block_structure(other)
            assert self.shape == other.shape, 'Dimension mismatch {} != {}'.format(
                self.shape, other.shape
            )
            assert (
                self.nblocks == other.nblocks
            ), 'Number of blocks mismatch {} != {}'.format(self.nblocks, other.nblocks)
            for idx, blk in enumerate(self):
                blk -= other.get_block(idx)
            return self
        elif type(other) == np.ndarray:
            assert self.shape == other.shape, 'Dimension mismatch {} != {}'.format(
                self.shape, other.shape
            )
            accum = 0
            for idx, blk in enumerate(self):
                nelements = self._brow_lengths[idx]
                blk -= other[accum : accum + nelements]
                accum += nelements
            return self
        else:
            raise NotImplementedError()

    def __imul__(self, other):
        # elementwise inplace multiplication to this BlockVector with other vector
        # supports multiplication with scalar, numpy.ndarray and BlockVectors
        assert_block_structure(self)
        if np.isscalar(other):
            for idx, blk in enumerate(self):
                blk *= other
            return self
        elif isinstance(other, BlockVector):
            assert_block_structure(other)
            assert self.shape == other.shape, 'Dimension mismatch {} != {}'.format(
                self.shape, other.shape
            )
            assert (
                self.nblocks == other.nblocks
            ), 'Number of blocks mismatch {} != {}'.format(self.nblocks, other.nblocks)
            for idx, blk in enumerate(self):
                blk *= other.get_block(idx)
            return self
        elif type(other) == np.ndarray:
            assert self.shape == other.shape, 'Dimension mismatch {} != {}'.format(
                self.shape, other.shape
            )
            accum = 0
            for idx, blk in enumerate(self):
                nelements = self._brow_lengths[idx]
                blk *= other[accum : accum + nelements]
                accum += nelements
            return self
        else:
            raise NotImplementedError()

    def __itruediv__(self, other):
        # elementwise inplace division to this BlockVector with other vector
        # supports division with scalar, numpy.ndarray and BlockVectors
        assert_block_structure(self)
        if np.isscalar(other):
            for idx, blk in enumerate(self):
                blk /= other
            return self
        elif isinstance(other, BlockVector):
            assert_block_structure(other)
            assert self.shape == other.shape, 'Dimension mismatch {} != {}'.format(
                self.shape, other.shape
            )
            assert (
                self.nblocks == other.nblocks
            ), 'Number of blocks mismatch {} != {}'.format(self.nblocks, other.nblocks)
            for idx, blk in enumerate(self):
                blk /= other.get_block(idx)
            return self
        elif type(other) == np.ndarray:
            assert self.shape == other.shape, 'Dimension mismatch {} != {}'.format(
                self.shape, other.shape
            )
            accum = 0
            for idx, blk in enumerate(self):
                nelements = self._brow_lengths[idx]
                blk /= other[accum : accum + nelements]
                accum += nelements
            return self
        else:
            raise NotImplementedError()

    def __div__(self, other):
        return self.__truediv__(other)

    def __rdiv__(self, other):
        return self.__rtruediv__(other)

    def __idiv__(self, other):
        return self.__itruediv__(other)

    def _print(self, indent):
        msg = ''
        for ndx, block in enumerate(self):
            if isinstance(block, BlockVector):
                msg += (
                    indent
                    + str(ndx)
                    + ': '
                    + block.__class__.__name__
                    + str(block.bshape)
                    + '\n'
                )
                msg += block._print(indent=indent + '   ')
            else:
                msg += (
                    indent
                    + str(ndx)
                    + ': '
                    + block.__class__.__name__
                    + str(block.shape)
                    + '\n'
                )
        return msg

    def __str__(self):
        return self._print(indent='')

    def __repr__(self):
        return '{}{}'.format(self.__class__.__name__, self.bshape)

    def get_block(self, key):
        """
        Access a block.

        Parameters
        ----------
        key: int
            This is the block index

        Returns
        -------
        block: np.ndarray or BlockVector
            The block corresponding to the index key.
        """
        return super(BlockVector, self).__getitem__(key)

    def set_block(self, key, value):
        """
        Set a block. The value can be a NumPy array or another BlockVector.

        Parameters
        ----------
        key: int
            This is the block index
        value:
            This is the block. It can be a NumPy array or another BlockVector.
        """
        assert -self.nblocks < key < self.nblocks, 'out of range'
        assert isinstance(value, np.ndarray) or isinstance(
            value, BaseBlockVector
        ), 'Blocks need to be numpy arrays or BlockVectors'
        assert value.ndim == 1, 'Blocks need to be 1D'

        if isinstance(value, BaseBlockVector):
            assert_block_structure(value)
        self._set_block_size(key, value.size)
        super(BlockVector, self).__setitem__(key, value)

    def _has_equal_structure(self, other):
        """
        Parameters
        ----------
        other: BlockVector

        Returns
        -------
        equal_structure: bool
            True if self and other have the same block structure (recursive). False otherwise.
        """
        if not isinstance(other, BlockVector):
            return False
        if self.nblocks != other.nblocks:
            return False
        for ndx, block1 in enumerate(self):
            block2 = other.get_block(ndx)
            if isinstance(block1, BlockVector):
                if not isinstance(block2, BlockVector):
                    return False
                if not block1._has_equal_structure(block2):
                    return False
            elif isinstance(block2, BlockVector):
                return False
        return True

    def __getitem__(self, item):
        # numpy: __getitem__[()] is identity
        if item.__class__ is tuple and not item:
            return self
        if not self._has_equal_structure(item):
            raise ValueError(
                'BlockVector.__getitem__ only accepts slices in the form of BlockVectors of the same structure'
            )
        res = BlockVector(self.nblocks)
        for ndx, block in self:
            res.set_block(ndx, block[item.get_block(ndx)])

    def __setitem__(self, key, value):
        if not (
            self._has_equal_structure(key)
            and (self._has_equal_structure(value) or np.isscalar(value))
        ):
            raise ValueError(
                'BlockVector.__setitem__ only accepts slices in the form of BlockVectors of the same structure'
            )
        if np.isscalar(value):
            for ndx, block in enumerate(self):
                block[key.get_block(ndx)] = value
        else:
            for ndx, block in enumerate(self):
                block[key.get_block(ndx)] = value.get_block(ndx)

    def _comparison_helper(self, other, operation):
        assert_block_structure(self)
        result = self.copy_structure()
        if isinstance(other, BlockVector):
            assert_block_structure(other)
            for ndx in range(self.nblocks):
                result.set_block(
                    ndx, operation(self.get_block(ndx), other.get_block(ndx))
                )
            return result
        elif isinstance(other, np.ndarray):
            assert self.shape == other.shape, 'Dimension mismatch {} != {}'.format(
                self.shape, other.shape
            )
            accum = 0
            for ndx in range(self.nblocks):
                result.set_block(
                    ndx,
                    operation(
                        self.get_block(ndx),
                        other[accum : accum + self.get_block_size(ndx)],
                    ),
                )
                accum += self.get_block_size(ndx)
            return result
        elif np.isscalar(other):
            for ndx in range(self.nblocks):
                result.set_block(ndx, operation(self.get_block(ndx), other))
            return result
        else:
            raise NotImplementedError('Operation not supported by BlockVector')

    def __le__(self, other):
        return self._comparison_helper(other, operator.le)

    def __lt__(self, other):
        return self._comparison_helper(other, operator.lt)

    def __ge__(self, other):
        return self._comparison_helper(other, operator.ge)

    def __gt__(self, other):
        return self._comparison_helper(other, operator.gt)

    def __eq__(self, other):
        return self._comparison_helper(other, operator.eq)

    def __ne__(self, other):
        return self._comparison_helper(other, operator.ne)

    def __neg__(self):
        # elementwise negate this BlockVector
        assert_block_structure(self)
        bv = BlockVector(self.nblocks)
        for bid in range(self.nblocks):
            bv.set_block(bid, self.get_block(bid).__neg__())
        return bv

    def __contains__(self, item):
        other = item
        assert_block_structure(self)
        if np.isscalar(other):
            contains = False
            for idx, blk in enumerate(self):
                if blk.__contains__(other):
                    return True
            return contains
        else:
            raise NotImplementedError()

    def __len__(self):
        return self.nblocks

    def pprint(self):
        """Prints BlockVector in pretty format"""
        msg = self.__repr__()
        msg += '\n'
        msg += self.__str__()
        print(msg)

    def toMPIBlockVector(self, rank_ownership, mpi_comm, assert_correct_owners=False):
        """
        Creates a parallel MPIBlockVector from this BlockVector

        Parameters
        ----------
        rank_ownership: array_like
            Array_like of size nblocks. Each entry defines ownership of each block.
            There are two types of ownership. Block that are owned by all processor,
            and blocks owned by a single processor. If a block is owned by all
            processors then its ownership is -1. Otherwise, if a block is owned by
            a single processor, then its ownership is equal to the rank of the
            processor.
        mpi_comm: MPI communicator
            An MPI communicator. Tyically MPI.COMM_WORLD

        """
        from pyomo.contrib.pynumero.sparse.mpi_block_vector import MPIBlockVector

        assert_block_structure(self)
        assert (
            len(rank_ownership) == self.nblocks
        ), 'rank_ownership must be of size {}'.format(self.nblocks)

        mpi_bv = MPIBlockVector(
            self.nblocks,
            rank_ownership,
            mpi_comm,
            assert_correct_owners=assert_correct_owners,
        )

        # populate blocks in the right spaces
        for bid in mpi_bv.owned_blocks:
            mpi_bv.set_block(bid, self.get_block(bid))

        return mpi_bv



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import os.path

from pyomo.common.collections import Bunch
from pyomo.dataportal.factory import DataManagerFactory
from pyomo.dataportal.process_data import _process_include


@DataManagerFactory.register("dat", "Pyomo data command file interface")
class PyomoDataCommands:
    def __init__(self):
        self._info = []
        self.options = Bunch()

    def available(self):
        return True

    def initialize(self, **kwds):
        self.filename = kwds.pop('filename')
        self.add_options(**kwds)

    def add_options(self, **kwds):
        self.options.update(kwds)

    def open(self):
        if self.filename is None:  # pragma:nocover
            raise IOError("No filename specified")
        if not os.path.exists(self.filename):  # pragma:nocover
            raise IOError("Cannot find file '%s'" % self.filename)

    def close(self):
        pass

    def read(self):
        """This function does nothing, since executing Pyomo data commands both
        reads and processes the data all at once.

        """
        pass

    def write(self, data):  # pragma:nocover
        """This function does nothing, because we cannot write to a ``*.dat`` file."""
        pass

    def process(self, model, data, default):
        """Read Pyomo data commands and process the data."""
        _process_include(['include', self.filename], model, data, default, self.options)

    def clear(self):
        self._info = []



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.contrib.piecewise.piecewise_linear_expression import (
    PiecewiseLinearExpression,
)
from pyomo.contrib.piecewise.piecewise_linear_function import PiecewiseLinearFunction

## register transformations
from pyomo.contrib.piecewise.transform.inner_representation_gdp import (
    InnerRepresentationGDPTransformation,
)
from pyomo.contrib.piecewise.transform.disaggregated_convex_combination import (
    DisaggregatedConvexCombinationTransformation,
)
from pyomo.contrib.piecewise.transform.outer_representation_gdp import (
    OuterRepresentationGDPTransformation,
)
from pyomo.contrib.piecewise.transform.multiple_choice import (
    MultipleChoiceTransformation,
)
from pyomo.contrib.piecewise.transform.reduced_inner_representation_gdp import (
    ReducedInnerRepresentationGDPTransformation,
)
from pyomo.contrib.piecewise.transform.convex_combination import (
    ConvexCombinationTransformation,
)
from pyomo.contrib.piecewise.transform.nonlinear_to_pwl import (
    DomainPartitioningMethod,
    NonlinearToPWL,
)
from pyomo.contrib.piecewise.transform.nested_inner_repn import (
    NestedInnerRepresentationGDPTransformation,
)
from pyomo.contrib.piecewise.transform.disaggregated_logarithmic import (
    DisaggregatedLogarithmicMIPTransformation,
)
from pyomo.contrib.piecewise.transform.incremental import IncrementalMIPTransformation
from pyomo.contrib.piecewise.triangulations import Triangulation



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.core.base.objective import Objective


def get_objective(block):
    obj = None
    for o in block.component_data_objects(
        Objective, descend_into=True, active=True, sort=True
    ):
        if obj is not None:
            raise ValueError('Multiple active objectives found')
        obj = o
    return obj



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.core.expr.logical_expr import BooleanExpression


class NoOverlapExpression(BooleanExpression):
    """
    Expression representing that none of the IntervalVars in a SequenceVar overlap
    (if they are scheduled)

    args:
        args (tuple): Child node of type SequenceVar
    """

    def nargs(self):
        return 1

    def _to_string(self, values, verbose, smap):
        return "no_overlap(%s)" % values[0]


class FirstInSequenceExpression(BooleanExpression):
    """
    Expression representing that the specified IntervalVar is the first in the
    sequence specified by SequenceVar (if it is scheduled)

    args:
        args (tuple): Child nodes, the first of type IntervalVar, the second of type
                      SequenceVar
    """

    def nargs(self):
        return 2

    def _to_string(self, values, verbose, smap):
        return "first_in(%s, %s)" % (values[0], values[1])


class LastInSequenceExpression(BooleanExpression):
    """
    Expression representing that the specified IntervalVar is the last in the
    sequence specified by SequenceVar (if it is scheduled)

    args:
        args (tuple): Child nodes, the first of type IntervalVar, the second of type
                      SequenceVar
    """

    def nargs(self):
        return 2

    def _to_string(self, values, verbose, smap):
        return "last_in(%s, %s)" % (values[0], values[1])


class BeforeInSequenceExpression(BooleanExpression):
    """
    Expression representing that one IntervalVar occurs before another in the
    sequence specified by the given SequenceVar (if both are scheduled)

    args:
        args (tuple): Child nodes, the IntervalVar that must be before, the
                      IntervalVar that must be after, and the SequenceVar
    """

    def nargs(self):
        return 3

    def _to_string(self, values, verbose, smap):
        return "before_in(%s, %s, %s)" % (values[0], values[1], values[2])


class PredecessorToExpression(BooleanExpression):
    """
    Expression representing that one IntervalVar is a direct predecessor to another
    in the sequence specified by the given SequenceVar (if both are scheduled)

    args:
        args (tuple): Child nodes, the predecessor IntervalVar, the successor
                      IntervalVar, and the SequenceVar
    """

    def nargs(self):
        return 3

    def _to_string(self, values, verbose, smap):
        return "predecessor_to(%s, %s, %s)" % (values[0], values[1], values[2])


def no_overlap(sequence_var):
    """
    Creates a new NoOverlapExpression

    Requires that none of the scheduled intervals in the SequenceVar overlap each other

    args:
        sequence_var: A SequenceVar
    """
    return NoOverlapExpression((sequence_var,))


def first_in_sequence(interval_var, sequence_var):
    """
    Creates a new FirstInSequenceExpression

    Requires that 'interval_var' be the first in the sequence specified by
    'sequence_var' if it is scheduled

    args:
        interval_var (IntervalVar): The activity that should be scheduled first
            if it is scheduled at all
        sequence_var (SequenceVar): The sequence of activities
    """
    return FirstInSequenceExpression((interval_var, sequence_var))


def last_in_sequence(interval_var, sequence_var):
    """
    Creates a new LastInSequenceExpression

    Requires that 'interval_var' be the last in the sequence specified by
    'sequence_var' if it is scheduled

    args:
        interval_var (IntervalVar): The activity that should be scheduled last
            if it is scheduled at all
        sequence_var (SequenceVar): The sequence of activities
    """

    return LastInSequenceExpression((interval_var, sequence_var))


def before_in_sequence(before_var, after_var, sequence_var):
    """
    Creates a new BeforeInSequenceExpression

    Requires that 'before_var' be scheduled to start before 'after_var' in the
    sequence specified bv 'sequence_var', if both are scheduled

    args:
        before_var (IntervalVar): The activity that should be scheduled earlier in
            the sequence
        after_var (IntervalVar): The activity that should be scheduled later in the
            sequence
        sequence_var (SequenceVar): The sequence of activities
    """
    return BeforeInSequenceExpression((before_var, after_var, sequence_var))


def predecessor_to(before_var, after_var, sequence_var):
    """
    Creates a new PredecessorToExpression

    Requires that 'before_var' be a direct predecessor to 'after_var' in the
    sequence specified by 'sequence_var', if both are scheduled

    args:
        before_var (IntervalVar): The activity that should be scheduled as the
            predecessor
        after_var (IntervalVar): The activity that should be scheduled as the
            successor
        sequence_var (SequenceVar): The sequence of activities
    """
    return PredecessorToExpression((before_var, after_var, sequence_var))



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################

import pyomo.common.unittest as unittest
import pytest

import pyomo.environ as pyo
import pyomo.dae as dae
import pyomo.contrib.mpc as mpc
from pyomo.contrib.mpc.data.scalar_data import ScalarData
from pyomo.contrib.mpc.data.interval_data import assert_disjoint_intervals, IntervalData


class TestIntervalData(unittest.TestCase):
    def _make_model(self):
        m = pyo.ConcreteModel()
        m.time = pyo.Set(initialize=[0.1 * i for i in range(11)])
        m.comp = pyo.Set(initialize=["A", "B"])
        m.var = pyo.Var(m.time, m.comp, initialize=1.0)
        return m

    def test_construct(self):
        m = self._make_model()
        intervals = [(0.0, 0.5), (0.5, 1.0)]
        data = {m.var[:, "A"]: [1.0, 2.0], m.var[:, "B"]: [3.0, 4.0]}
        interval_data = IntervalData(data, intervals)

        self.assertEqual(
            interval_data.get_data(),
            {pyo.ComponentUID(key): val for key, val in data.items()},
        )
        self.assertEqual(intervals, interval_data.get_intervals())

    def test_eq(self):
        m = self._make_model()
        intervals = [(0.0, 0.5), (0.5, 1.0)]
        data = {m.var[:, "A"]: [1.0, 2.0], m.var[:, "B"]: [3.0, 4.0]}
        interval_data_1 = IntervalData(data, intervals)

        data = {m.var[:, "A"]: [1.0, 2.0], m.var[:, "B"]: [3.0, 4.0]}
        interval_data_2 = IntervalData(data, intervals)

        self.assertEqual(interval_data_1, interval_data_2)

        data = {m.var[:, "A"]: [1.0, 3.0], m.var[:, "B"]: [3.0, 4.0]}
        interval_data_3 = IntervalData(data, intervals)

        self.assertNotEqual(interval_data_1, interval_data_3)

    def test_get_data_at_indices_multiple(self):
        m = self._make_model()
        intervals = [(0.0, 0.2), (0.2, 0.5), (0.5, 1.0)]
        data = {m.var[:, "A"]: [1.0, 2.0, 3.0], m.var[:, "B"]: [4.0, 5.0, 6.0]}
        interval_data = IntervalData(data, intervals)
        data = interval_data.get_data_at_interval_indices([0, 2])

        pred_data = IntervalData(
            {m.var[:, "A"]: [1.0, 3.0], m.var[:, "B"]: [4.0, 6.0]},
            [(0.0, 0.2), (0.5, 1.0)],
        )
        self.assertEqual(pred_data, data)

    def test_get_data_at_indices_singleton(self):
        m = self._make_model()
        intervals = [(0.0, 0.2), (0.2, 0.5), (0.5, 1.0)]
        data = {m.var[:, "A"]: [1.0, 2.0, 3.0], m.var[:, "B"]: [4.0, 5.0, 6.0]}
        interval_data = IntervalData(data, intervals)
        data = interval_data.get_data_at_interval_indices(1)
        pred_data = ScalarData({m.var[:, "A"]: 2.0, m.var[:, "B"]: 5.0})
        self.assertEqual(data, pred_data)

    def test_get_data_at_time_scalar(self):
        m = self._make_model()
        intervals = [(0.0, 0.2), (0.2, 0.5), (0.5, 1.0)]
        data = {m.var[:, "A"]: [1.0, 2.0, 3.0], m.var[:, "B"]: [4.0, 5.0, 6.0]}
        interval_data = IntervalData(data, intervals)

        data = interval_data.get_data_at_time(0.1)
        pred_data = ScalarData({m.var[:, "A"]: 1.0, m.var[:, "B"]: 4.0})
        self.assertEqual(data, pred_data)

        # Default is to allow time points outside of intervals
        # (finds the nearest interval)
        data = interval_data.get_data_at_time(1.1)
        pred_data = ScalarData({m.var[:, "A"]: 3.0, m.var[:, "B"]: 6.0})
        self.assertEqual(data, pred_data)

        msg = "Time point.*not found"
        with self.assertRaisesRegex(RuntimeError, msg):
            data = interval_data.get_data_at_time(1.1, tolerance=1e-3)

        # If a point on an interval boundary is supplied, default is to
        # use value on left.
        data = interval_data.get_data_at_time(0.5)
        pred_data = ScalarData({m.var[:, "A"]: 2.0, m.var[:, "B"]: 5.0})
        self.assertEqual(data, pred_data)

        data = interval_data.get_data_at_time(0.5, prefer_left=False)
        pred_data = ScalarData({m.var[:, "A"]: 3.0, m.var[:, "B"]: 6.0})
        self.assertEqual(data, pred_data)

    def test_to_serializable(self):
        m = self._make_model()
        intervals = [(0.0, 0.2), (0.2, 0.5), (0.5, 1.0)]
        data = {m.var[:, "A"]: [1.0, 2.0, 3.0], m.var[:, "B"]: [4.0, 5.0, 6.0]}
        interval_data = IntervalData(data, intervals)
        json_data = interval_data.to_serializable()
        self.assertEqual(
            json_data,
            (
                {"var[*,A]": [1.0, 2.0, 3.0], "var[*,B]": [4.0, 5.0, 6.0]},
                [(0.0, 0.2), (0.2, 0.5), (0.5, 1.0)],
            ),
        )

    def test_concatenate(self):
        m = self._make_model()
        intervals = [(0.0, 0.2), (0.2, 0.5), (0.5, 1.0)]
        data = {m.var[:, "A"]: [1.0, 2.0, 3.0], m.var[:, "B"]: [4.0, 5.0, 6.0]}
        interval_data_1 = IntervalData(data, intervals)

        intervals = [(1.0, 1.5), (2.0, 3.0)]
        data = {m.var[:, "A"]: [7.0, 8.0], m.var[:, "B"]: [9.0, 10.0]}
        interval_data_2 = IntervalData(data, intervals)

        interval_data_1.concatenate(interval_data_2)

        new_intervals = [(0.0, 0.2), (0.2, 0.5), (0.5, 1.0), (1.0, 1.5), (2.0, 3.0)]
        new_values = {
            m.var[:, "A"]: [1.0, 2.0, 3.0, 7.0, 8.0],
            m.var[:, "B"]: [4.0, 5.0, 6.0, 9.0, 10.0],
        }
        new_data = IntervalData(new_values, new_intervals)
        self.assertEqual(interval_data_1, new_data)

    def test_shift_time_points(self):
        m = self._make_model()
        intervals = [(0.0, 0.2), (0.2, 0.5), (0.5, 1.0)]
        data = {m.var[:, "A"]: [1.0, 2.0, 3.0], m.var[:, "B"]: [4.0, 5.0, 6.0]}
        interval_data = IntervalData(data, intervals)
        interval_data.shift_time_points(1.0)

        intervals = [(1.0, 1.2), (1.2, 1.5), (1.5, 2.0)]
        data = {m.var[:, "A"]: [1.0, 2.0, 3.0], m.var[:, "B"]: [4.0, 5.0, 6.0]}
        new_interval_data = IntervalData(data, intervals)
        self.assertEqual(interval_data, new_interval_data)

    def test_extract_variables(self):
        m = self._make_model()
        intervals = [(0.0, 0.2), (0.2, 0.5), (0.5, 1.0)]
        data = {m.var[:, "A"]: [1.0, 2.0, 3.0], m.var[:, "B"]: [4.0, 5.0, 6.0]}
        interval_data = IntervalData(data, intervals, time_set=m.time)
        new_data = interval_data.extract_variables([m.var[:, "B"]])
        value_dict = {m.var[:, "B"]: [4.0, 5.0, 6.0]}
        pred_data = IntervalData(value_dict, intervals)
        self.assertEqual(new_data, pred_data)

    def test_extract_variables_exception(self):
        m = self._make_model()
        intervals = [(0.0, 0.2), (0.2, 0.5), (0.5, 1.0)]
        data = {m.var[:, "A"]: [1.0, 2.0, 3.0], m.var[:, "B"]: [4.0, 5.0, 6.0]}
        interval_data = IntervalData(data, intervals, time_set=m.time)
        msg = "only accepts a list or tuple"
        with self.assertRaisesRegex(TypeError, msg):
            new_data = interval_data.extract_variables(m.var[:, "B"])


class TestAssertDisjoint(unittest.TestCase):
    def test_disjoint(self):
        intervals = [(0, 1), (1, 2)]
        assert_disjoint_intervals(intervals)

        intervals = [(2, 3), (0, 1)]
        assert_disjoint_intervals(intervals)

        intervals = [(0, 1), (1, 1)]
        assert_disjoint_intervals(intervals)

    def test_backwards_endpoints(self):
        intervals = [(0, 1), (3, 2)]
        msg = "Lower endpoint of interval is higher"
        with self.assertRaisesRegex(RuntimeError, msg):
            assert_disjoint_intervals(intervals)

    def test_not_disjoint(self):
        intervals = [(0, 2), (1, 3)]
        msg = "are not disjoint"
        with self.assertRaisesRegex(RuntimeError, msg):
            assert_disjoint_intervals(intervals)


class TestLoadInputs(unittest.TestCase):
    def make_model(self):
        m = pyo.ConcreteModel()
        m.time = dae.ContinuousSet(initialize=[0, 1, 2, 3, 4, 5, 6])
        m.v = pyo.Var(m.time, initialize=0)
        return m

    def test_load_inputs_some_time(self):
        m = self.make_model()
        interface = mpc.DynamicModelInterface(m, m.time)
        inputs = mpc.IntervalData({"v": [1.0]}, [(2, 4)])
        interface.load_data(inputs)

        for t in m.time:
            # Note that by default, the left endpoint is not loaded.
            if t == 3 or t == 4:
                self.assertEqual(m.v[t].value, 1.0)
            else:
                self.assertEqual(m.v[t].value, 0.0)

    def test_load_inputs_some_time_include_endpoints(self):
        m = self.make_model()
        interface = mpc.DynamicModelInterface(m, m.time)
        inputs = mpc.IntervalData({"v": [1.0]}, [(2, 4)])

        # Default is to exclude right and include left
        interface.load_data(inputs, exclude_left_endpoint=False)

        for t in m.time:
            if t == 2 or t == 3 or t == 4:
                self.assertEqual(m.v[t].value, 1.0)
            else:
                self.assertEqual(m.v[t].value, 0.0)

    def test_load_inputs_some_time_exclude_endpoints(self):
        m = self.make_model()
        interface = mpc.DynamicModelInterface(m, m.time)
        inputs = mpc.IntervalData({"v": [1.0]}, [(2, 4)])

        # Default is to exclude right and include left
        interface.load_data(inputs, exclude_right_endpoint=True)

        for t in m.time:
            if t == 3:
                self.assertEqual(m.v[t].value, 1.0)
            else:
                self.assertEqual(m.v[t].value, 0.0)

    def test_load_inputs_all_time_default(self):
        m = self.make_model()
        interface = mpc.DynamicModelInterface(m, m.time)
        inputs = mpc.IntervalData({"v": [1.0, 2.0]}, [(0, 3), (3, 6)])
        interface.load_data(inputs)
        for t in m.time:
            if t == 0:
                self.assertEqual(m.v[t].value, 0.0)
            elif t <= 3:
                self.assertEqual(m.v[t].value, 1.0)
            else:
                self.assertEqual(m.v[t].value, 2.0)

    def test_load_inputs_all_time_prefer_right(self):
        m = self.make_model()
        interface = mpc.DynamicModelInterface(m, m.time)
        inputs = mpc.IntervalData({"v": [1.0, 2.0]}, [(0, 3), (3, 6)])
        interface.load_data(inputs, prefer_left=False)
        for t in m.time:
            if t < 3:
                self.assertEqual(m.v[t].value, 1.0)
            elif t == 6:
                # By default, preferring intervals to the right of time
                # points will exclude the right endpoints of intervals.
                self.assertEqual(m.v[t].value, 0.0)
            else:
                self.assertEqual(m.v[t].value, 2.0)

    def test_load_inputs_all_time_prefer_right_dont_exclude(self):
        m = self.make_model()
        interface = mpc.DynamicModelInterface(m, m.time)
        inputs = mpc.IntervalData({"v": [1.0, 2.0]}, [(0, 3), (3, 6)])
        interface.load_data(inputs, prefer_left=False, exclude_right_endpoint=False)
        # Note that all time points have been set.
        for t in m.time:
            if t < 3:
                self.assertEqual(m.v[t].value, 1.0)
            else:
                self.assertEqual(m.v[t].value, 2.0)

    def load_inputs_invalid_time(self):
        m = self.make_model()
        interface = mpc.DynamicModelInterface(m, m.time)
        inputs = mpc.IntervalData({"v": [1.0, 2.0, 3.0]}, [(0, 3), (3, 6), (6, 9)])
        interface.load_data(inputs)
        for t in m.time:
            if t == 0:
                self.assertEqual(m.v[t].value, 0.0)
            elif t <= 3:
                self.assertEqual(m.v[t].value, 1.0)
            else:
                self.assertEqual(m.v[t].value, 2.0)

    def load_inputs_exception(self):
        m = self.make_model()
        interface = mpc.DynamicModelInterface(m, m.time)
        inputs = {"_v": {(0, 3): 1.0, (3, 6): 2.0, (6, 9): 3.0}}
        inputs = mpc.IntervalData({"_v": [1.0, 2.0, 3.0]}, [(0, 3), (3, 6), (6, 9)])
        with self.assertRaisesRegex(RuntimeError, "Cannot find"):
            interface.load_data(inputs)


class TestIntervalFromTimeSeries(unittest.TestCase):
    def test_singleton(self):
        name = "name"
        series = mpc.TimeSeriesData({name: [2.0]}, [1.0])
        interval = mpc.data.convert.series_to_interval(series)
        self.assertEqual(interval, IntervalData({name: [2.0]}, [(1.0, 1.0)]))

    def test_empty(self):
        name = "name"
        series = mpc.TimeSeriesData({name: []}, [])
        interval = mpc.data.convert.series_to_interval(series)
        self.assertEqual(interval, mpc.IntervalData({name: []}, []))

    def test_interval_from_series(self):
        name = "name"
        series = mpc.TimeSeriesData({name: [4.0, 5.0, 6.0]}, [1, 2, 3])
        interval = mpc.data.convert.series_to_interval(series)
        self.assertEqual(
            interval, mpc.IntervalData({name: [5.0, 6.0]}, [(1, 2), (2, 3)])
        )

    def test_use_left_endpoint(self):
        name = "name"
        series = mpc.TimeSeriesData({name: [4.0, 5.0, 6.0]}, [1, 2, 3])
        interval = mpc.data.convert.series_to_interval(series, use_left_endpoints=True)
        self.assertEqual(
            interval, mpc.IntervalData({name: [4.0, 5.0]}, [(1, 2), (2, 3)])
        )


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.version import version_info, __version__

#
# Load solver functionality
#
import pyomo.environ
from pyomo.opt import SolverFactory, SolverStatus, TerminationCondition

#
# Define the modeling namespace
#
from pyomo.common.collections import ComponentMap, ComponentSet
from pyomo.core.expr import (
    numvalue,
    numeric_expr,
    boolean_value,
    logical_expr,
    calculus,
    symbol_map,
    expr_errors,
    visitor,
    sympy_tools,
    taylor_series,
    expr_common,
    cnf_walker,
    template_expr,
)


from pyomo.core.expr.numvalue import (
    value,
    is_constant,
    is_fixed,
    is_variable_type,
    is_potentially_variable,
    NumericValue,
    ZeroConstant,
    native_numeric_types,
    native_types,
    polynomial_degree,
)

from pyomo.core.expr.boolean_value import BooleanValue

from pyomo.core.expr import (
    linear_expression,
    nonlinear_expression,
    land,
    lor,
    equivalent,
    exactly,
    atleast,
    atmost,
    implies,
    lnot,
    xor,
    inequality,
    log,
    log10,
    sin,
    cos,
    tan,
    cosh,
    sinh,
    tanh,
    asin,
    acos,
    atan,
    exp,
    sqrt,
    asinh,
    acosh,
    atanh,
    ceil,
    floor,
    Expr_if,
)

from pyomo.core.expr.calculus.derivatives import differentiate
from pyomo.core.expr.taylor_series import taylor_series_expansion
from pyomo.kernel.util import generate_names, preorder_traversal, pprint
from pyomo.core.kernel.variable import (
    variable,
    variable_tuple,
    variable_list,
    variable_dict,
)
from pyomo.core.kernel.constraint import (
    constraint,
    linear_constraint,
    constraint_tuple,
    constraint_list,
    constraint_dict,
)
from pyomo.core.kernel.matrix_constraint import matrix_constraint
import pyomo.core.kernel.conic as conic
from pyomo.core.kernel.parameter import (
    parameter,
    functional_value,
    parameter_tuple,
    parameter_list,
    parameter_dict,
)
from pyomo.core.kernel.expression import (
    noclone,
    expression,
    data_expression,
    expression_tuple,
    expression_list,
    expression_dict,
)
from pyomo.core.kernel.objective import (
    maximize,
    minimize,
    objective,
    objective_tuple,
    objective_list,
    objective_dict,
)
from pyomo.core.kernel.sos import sos, sos1, sos2, sos_tuple, sos_list, sos_dict
from pyomo.core.kernel.suffix import (
    suffix,
    suffix_dict,
    export_suffix_generator,
    import_suffix_generator,
    local_suffix_generator,
    suffix_generator,
)
from pyomo.core.kernel.block import block, block_tuple, block_list, block_dict
from pyomo.core.kernel.piecewise_library.transforms import piecewise
from pyomo.core.kernel.piecewise_library.transforms_nd import piecewise_nd
from pyomo.core.kernel.set_types import RealSet, IntegerSet, BooleanSet
from pyomo.environ import (
    Reals,
    PositiveReals,
    NonPositiveReals,
    NegativeReals,
    NonNegativeReals,
    PercentFraction,
    UnitInterval,
    Integers,
    PositiveIntegers,
    NonPositiveIntegers,
    NegativeIntegers,
    NonNegativeIntegers,
    Boolean,
    Binary,
    RealInterval,
    IntegerInterval,
)

#
# allow the use of standard kernel modeling components
# as the ctype argument for the general iterator method
#

from pyomo.core.kernel.base import _convert_ctype

_convert_ctype[block] = pyomo.core.kernel.block.IBlock
_convert_ctype[variable] = pyomo.core.kernel.variable.IVariable
_convert_ctype[constraint] = pyomo.core.kernel.constraint.IConstraint
_convert_ctype[parameter] = pyomo.core.kernel.parameter.IParameter
_convert_ctype[expression] = pyomo.core.kernel.expression.IExpression
_convert_ctype[objective] = pyomo.core.kernel.objective.IObjective
_convert_ctype[sos] = pyomo.core.kernel.sos.ISOS
_convert_ctype[suffix] = pyomo.core.kernel.suffix.ISuffix
del _convert_ctype

#
#
# Hacks needed for this interface to work with Pyomo solvers
#
#
from pyomo.core.kernel.base import _convert_ctype, _kernel_ctype_backmap

#
# Set up mappings between AML and Kernel ctypes
#
_convert_ctype[pyomo.environ.Block] = pyomo.core.kernel.block.IBlock
_convert_ctype[pyomo.environ.Var] = pyomo.core.kernel.variable.IVariable
_convert_ctype[pyomo.environ.Constraint] = pyomo.core.kernel.constraint.IConstraint
_convert_ctype[pyomo.environ.Param] = pyomo.core.kernel.parameter.IParameter
_convert_ctype[pyomo.environ.Expression] = pyomo.core.kernel.expression.IExpression
_convert_ctype[pyomo.environ.Objective] = pyomo.core.kernel.objective.IObjective
_convert_ctype[pyomo.environ.SOSConstraint] = pyomo.core.kernel.sos.ISOS
_convert_ctype[pyomo.environ.Suffix] = pyomo.core.kernel.suffix.ISuffix

#
# Set up back mappings from Kernel back to AML ctypes
#
_kernel_ctype_backmap.update(
    {
        v: k
        for k, v in _convert_ctype.items()
        if not issubclass(k, pyomo.core.kernel.base.ICategorizedObject)
    }
)

del _convert_ctype
del _kernel_ctype_backmap

#
# Now cleanup the namespace a bit
#

import pyomo.core.kernel.piecewise_library.util as piecewise_util

del util
del pyomo

#
# Ducktyping to work with a solver interfaces. Ideally,
# everything below here could be deleted one day.
#
from pyomo.core.kernel.heterogeneous_container import (
    heterogeneous_containers,
    IHeterogeneousContainer,
)


def _component_data_objects(self, *args, **kwds):
    # this is not yet handled
    kwds.pop('sort', None)
    if 'active' not in kwds:
        kwds['active'] = None
    yield from self.components(*args, **kwds)


IHeterogeneousContainer.component_data_objects = _component_data_objects
del _component_data_objects


def _component_objects(self, *args, **kwds):
    # this is not yet handled
    kwds.pop('sort', None)
    # not handled
    assert kwds.pop('descent_order', None) is None
    active = kwds.pop('active', None)
    descend_into = kwds.pop('descend_into', True)
    for item in heterogeneous_containers(
        self, active=active, descend_into=descend_into
    ):
        yield from item.children(*args, **kwds)


IHeterogeneousContainer.component_objects = _component_objects
del _component_objects
del IHeterogeneousContainer


def _block_data_objects(self, **kwds):
    # this is not yet handled
    kwds.pop('sort', None)
    active = kwds.get("active", None)
    assert active in (None, True)
    # if not active, then nothing below is active
    if (active is not None) and (not self.active):
        return
    yield self
    yield from self.components(ctype=self.ctype, **kwds)


block.block_data_objects = _block_data_objects
del _block_data_objects


# Note sure where this gets used or why we need it
def _valid_problem_types(self):
    import pyomo.opt

    return [pyomo.opt.base.ProblemFormat.pyomo]


block.valid_problem_types = _valid_problem_types
del _valid_problem_types

# update the reserved block attributes now that
# new hacked methods have been placed on blocks
block._refresh_block_reserved_words()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#

import pyomo.common.unittest as unittest

from pyomo.environ import (
    ConcreteModel,
    AbstractModel,
    RangeSet,
    Param,
    Var,
    Set,
    value,
    Integers,
)
import pyomo.core.expr as EXPR
from pyomo.core.expr.template_expr import (
    IndexTemplate,
    TemplateExpressionError,
    CallExpression,
    GetAttrExpression,
    _GetItemIndexer,
    resolve_template,
    templatize_constraint,
    substitute_template_expression,
    substitute_getitem_with_param,
    substitute_template_with_value,
)


class TestTemplateExpressions(unittest.TestCase):
    def setUp(self):
        self.m = m = ConcreteModel()
        m.I = RangeSet(1, 9)
        m.J = RangeSet(10, 19)
        m.x = Var(m.I, initialize=lambda m, i: i + 1)
        m.P = Param(m.I, initialize=lambda m, i: 10 - i, mutable=True)
        m.p = Param(m.I, m.J, initialize=lambda m, i, j: 100 * i + j)
        m.s = Set(m.I, initialize=lambda m, i: range(i))

    def test_nonTemplates(self):
        m = self.m
        self.assertIs(resolve_template(m.x[1]), m.x[1])
        e = m.x[1] + m.x[2]
        self.assertIs(resolve_template(e), e)

    def test_IndexTemplate(self):
        m = self.m
        i = IndexTemplate(m.I)
        with self.assertRaisesRegex(
            TemplateExpressionError, "Evaluating uninitialized IndexTemplate"
        ):
            value(i)

        self.assertEqual(str(i), "{I}")

        i.set_value(5)
        self.assertEqual(value(i), 5)
        self.assertIs(resolve_template(i), 5)

    def test_template_scalar(self):
        m = self.m
        t = IndexTemplate(m.I)
        e = m.x[t]
        self.assertIs(type(e), EXPR.Numeric_GetItemExpression)
        self.assertEqual(e.args, (m.x, t))
        self.assertFalse(e.is_constant())
        self.assertFalse(e.is_fixed())
        self.assertEqual(e.polynomial_degree(), 1)
        self.assertEqual(str(e), "x[{I}]")
        t.set_value(5)
        v = e()
        self.assertIn(type(v), (int, float))
        self.assertEqual(v, 6)
        self.assertIs(resolve_template(e), m.x[5])
        t.set_value()

        e = m.p[t, 10]
        self.assertIs(type(e), EXPR.NPV_Numeric_GetItemExpression)
        self.assertEqual(e.args, (m.p, t, 10))
        self.assertFalse(e.is_constant())
        self.assertTrue(e.is_fixed())
        self.assertEqual(e.polynomial_degree(), 0)
        self.assertEqual(str(e), "p[{I},10]")
        t.set_value(5)
        v = e()
        self.assertIn(type(v), (int, float))
        self.assertEqual(v, 510)
        self.assertIs(resolve_template(e), m.p[5, 10])
        t.set_value()

        e = m.p[5, t]
        self.assertIs(type(e), EXPR.NPV_Numeric_GetItemExpression)
        self.assertEqual(e.args, (m.p, 5, t))
        self.assertFalse(e.is_constant())
        self.assertTrue(e.is_fixed())
        self.assertEqual(e.polynomial_degree(), 0)
        self.assertEqual(str(e), "p[5,{I}]")
        t.set_value(10)
        v = e()
        self.assertIn(type(v), (int, float))
        self.assertEqual(v, 510)
        self.assertIs(resolve_template(e), m.p[5, 10])
        t.set_value()

    def test_template_scalar_with_set(self):
        m = self.m
        t = IndexTemplate(m.I)
        e = m.s[t]
        self.assertIs(type(e), EXPR.NPV_Structural_GetItemExpression)
        self.assertEqual(e.args, (m.s, t))
        self.assertFalse(e.is_constant())
        self.assertTrue(e.is_fixed())
        # Generic templates generate CallExpression objects
        ee = e.polynomial_degree()
        self.assertIs(type(ee), EXPR.CallExpression)
        t.set_value(1)
        # Note that structural expressions do not implement polynomial_degree
        with self.assertRaisesRegex(
            AttributeError,
            "'InsertionOrderSetData' object has " "no attribute 'polynomial_degree'",
        ):
            e.polynomial_degree()
        self.assertEqual(str(e), "s[{I}]")
        t.set_value(5)
        v = e()
        self.assertIs(v, m.s[5])
        self.assertIs(resolve_template(e), m.s[5])
        t.set_value()

    def test_template_operation(self):
        m = self.m
        t = IndexTemplate(m.I)
        e = m.x[t + m.P[5]]
        self.assertIs(type(e), EXPR.Numeric_GetItemExpression)
        self.assertEqual(e.nargs(), 2)
        self.assertIs(e.arg(0), m.x)
        self.assertIsInstance(e.arg(1), EXPR.SumExpressionBase)
        self.assertIs(e.arg(1).arg(0), t)
        self.assertIs(e.arg(1).arg(1), m.P[5])
        self.assertEqual(str(e), "x[{I} + P[5]]")

    def test_nested_template_operation(self):
        m = self.m
        t = IndexTemplate(m.I)
        e = m.x[t + m.P[t + 1]]
        self.assertIs(type(e), EXPR.Numeric_GetItemExpression)
        self.assertEqual(e.nargs(), 2)
        self.assertIs(e.arg(0), m.x)
        self.assertIsInstance(e.arg(1), EXPR.SumExpressionBase)
        self.assertIs(e.arg(1).arg(0), t)
        self.assertIs(type(e.arg(1).arg(1)), EXPR.NPV_Numeric_GetItemExpression)
        self.assertIsInstance(e.arg(1).arg(1).arg(1), EXPR.SumExpressionBase)
        self.assertIs(e.arg(1).arg(1).arg(1).arg(0), t)
        self.assertEqual(str(e), "x[{I} + P[{I} + 1]]")

    def test_block_templates(self):
        m = ConcreteModel()
        m.T = RangeSet(3)

        @m.Block(m.T)
        def b(b, i):
            b.x = Var(initialize=i)

            @b.Block(m.T)
            def bb(bb, j):
                bb.I = RangeSet(i * j)
                bb.y = Var(bb.I, initialize=lambda m, i: i)

        t = IndexTemplate(m.T)
        e = m.b[t].x
        self.assertIs(type(e), EXPR.Numeric_GetAttrExpression)
        self.assertEqual(e.nargs(), 2)
        self.assertIs(type(e.arg(0)), EXPR.NPV_Structural_GetItemExpression)
        self.assertIs(e.arg(0).arg(0), m.b)
        self.assertEqual(e.arg(0).nargs(), 2)
        self.assertIs(e.arg(0).arg(1), t)
        self.assertEqual(str(e), "b[{T}].x")
        t.set_value(2)
        v = e()
        self.assertIn(type(v), (int, float))
        self.assertEqual(v, 2)
        self.assertIs(resolve_template(e), m.b[2].x)
        t.set_value()

        e = m.b[t].bb[t].y[1]
        self.assertIs(type(e), EXPR.Numeric_GetItemExpression)
        self.assertEqual(e.nargs(), 2)
        self.assertEqual(str(e), "b[{T}].bb[{T}].y[1]")
        t.set_value(2)
        v = e()
        self.assertIn(type(v), (int, float))
        self.assertEqual(v, 1)
        self.assertIs(resolve_template(e), m.b[2].bb[2].y[1])

    def test_template_name(self):
        m = self.m
        t = IndexTemplate(m.I)

        E = m.x[t + m.P[1 + t]] + m.P[1]
        self.assertEqual(str(E), "x[{I} + P[1 + {I}]] + P[1]")

        E = m.x[t + m.P[1 + t] ** 2.0] ** 2.0 + m.P[1]
        self.assertEqual(str(E), "x[{I} + P[1 + {I}]**2.0]**2.0 + P[1]")

    def test_template_in_expression(self):
        m = self.m
        t = IndexTemplate(m.I)

        E = m.x[t + m.P[t + 1]] + m.P[1]
        self.assertIsInstance(E, EXPR.SumExpressionBase)
        e = E.arg(0)
        self.assertIs(type(e), EXPR.Numeric_GetItemExpression)
        self.assertEqual(e.nargs(), 2)
        self.assertIs(e.arg(0), m.x)
        self.assertIsInstance(e.arg(1), EXPR.SumExpressionBase)
        self.assertIs(e.arg(1).arg(0), t)
        self.assertIs(type(e.arg(1).arg(1)), EXPR.NPV_Numeric_GetItemExpression)
        self.assertIsInstance(e.arg(1).arg(1).arg(1), EXPR.SumExpressionBase)
        self.assertIs(e.arg(1).arg(1).arg(1).arg(0), t)

        E = m.P[1] + m.x[t + m.P[t + 1]]
        self.assertIsInstance(E, EXPR.SumExpressionBase)
        e = E.arg(1)
        self.assertIs(type(e), EXPR.Numeric_GetItemExpression)
        self.assertEqual(e.nargs(), 2)
        self.assertIs(e.arg(0), m.x)
        self.assertIsInstance(e.arg(1), EXPR.SumExpressionBase)
        self.assertIs(e.arg(1).arg(0), t)
        self.assertIs(type(e.arg(1).arg(1)), EXPR.NPV_Numeric_GetItemExpression)
        self.assertIsInstance(e.arg(1).arg(1).arg(1), EXPR.SumExpressionBase)
        self.assertIs(e.arg(1).arg(1).arg(1).arg(0), t)

        E = m.x[t + m.P[t + 1]] + 1
        self.assertIsInstance(E, EXPR.SumExpressionBase)
        e = E.arg(0)
        self.assertIs(type(e), EXPR.Numeric_GetItemExpression)
        self.assertEqual(e.nargs(), 2)
        self.assertIs(e.arg(0), m.x)
        self.assertIsInstance(e.arg(1), EXPR.SumExpressionBase)
        self.assertIs(e.arg(1).arg(0), t)
        self.assertIs(type(e.arg(1).arg(1)), EXPR.NPV_Numeric_GetItemExpression)
        self.assertIsInstance(e.arg(1).arg(1).arg(1), EXPR.SumExpressionBase)
        self.assertIs(e.arg(1).arg(1).arg(1).arg(0), t)

        E = 1 + m.x[t + m.P[t + 1]]
        self.assertIsInstance(E, EXPR.SumExpressionBase)
        e = E.arg(E.nargs() - 1)
        self.assertIs(type(e), EXPR.Numeric_GetItemExpression)
        self.assertEqual(e.nargs(), 2)
        self.assertIs(e.arg(0), m.x)
        self.assertIsInstance(e.arg(1), EXPR.SumExpressionBase)
        self.assertIs(e.arg(1).arg(0), t)
        self.assertIs(type(e.arg(1).arg(1)), EXPR.NPV_Numeric_GetItemExpression)
        self.assertIsInstance(e.arg(1).arg(1).arg(1), EXPR.SumExpressionBase)
        self.assertIs(e.arg(1).arg(1).arg(1).arg(0), t)

    def test_clone(self):
        m = self.m
        t = IndexTemplate(m.I)

        E_base = m.x[t + m.P[t + 1]] + m.P[1]
        E = E_base.clone()
        self.assertIsInstance(E, EXPR.SumExpressionBase)
        e = E.arg(0)
        self.assertIs(type(e), EXPR.Numeric_GetItemExpression)
        self.assertIsNot(e, E_base.arg(0))
        self.assertEqual(e.nargs(), 2)
        self.assertIs(e.arg(0), m.x)
        self.assertIsInstance(e.arg(1), EXPR.SumExpressionBase)
        self.assertIs(e.arg(1).arg(0), t)
        self.assertIs(type(e.arg(1).arg(1)), EXPR.NPV_Numeric_GetItemExpression)
        self.assertIs(type(e.arg(1).arg(1)), type(E_base.arg(0).arg(1).arg(1)))
        self.assertIsNot(e.arg(1).arg(1), E_base.arg(0).arg(1).arg(1))
        self.assertIsInstance(e.arg(1).arg(1).arg(1), EXPR.SumExpressionBase)
        self.assertIs(e.arg(1).arg(1).arg(1).arg(0), t)

        E_base = m.P[1] + m.x[t + m.P[t + 1]]
        E = E_base.clone()
        self.assertTrue(isinstance(E, EXPR.SumExpressionBase))
        e = E.arg(1)
        self.assertIs(type(e), EXPR.Numeric_GetItemExpression)
        self.assertIsNot(e, E_base.arg(0))
        self.assertEqual(e.nargs(), 2)
        self.assertIs(e.arg(0), m.x)
        self.assertIsInstance(e.arg(1), EXPR.SumExpressionBase)
        self.assertIs(e.arg(1).arg(0), t)
        self.assertIs(type(e.arg(1).arg(1)), EXPR.NPV_Numeric_GetItemExpression)
        self.assertIs(type(e.arg(1).arg(1)), type(E_base.arg(1).arg(1).arg(1)))
        self.assertIsNot(e.arg(1).arg(1), E_base.arg(1).arg(1).arg(1))
        self.assertIsInstance(e.arg(1).arg(1).arg(1), EXPR.SumExpressionBase)
        self.assertIs(e.arg(1).arg(1).arg(1).arg(0), t)

        E_base = m.x[t + m.P[t + 1]] + 1
        E = E_base.clone()
        self.assertIsInstance(E, EXPR.SumExpressionBase)
        e = E.arg(0)
        self.assertIs(type(e), EXPR.Numeric_GetItemExpression)
        self.assertIsNot(e, E_base.arg(0))
        self.assertEqual(e.nargs(), 2)
        self.assertIs(e.arg(0), m.x)
        self.assertIsInstance(e.arg(1), EXPR.SumExpressionBase)
        self.assertIs(e.arg(1).arg(0), t)
        self.assertIs(type(e.arg(1).arg(1)), EXPR.NPV_Numeric_GetItemExpression)
        self.assertIs(type(e.arg(1).arg(1)), type(E_base.arg(0).arg(1).arg(1)))
        self.assertIsNot(e.arg(1).arg(1), E_base.arg(0).arg(1).arg(1))
        self.assertIsInstance(e.arg(1).arg(1).arg(1), EXPR.SumExpressionBase)
        self.assertIs(e.arg(1).arg(1).arg(1).arg(0), t)

        E_base = 1 + m.x[t + m.P[t + 1]]
        E = E_base.clone()
        self.assertIsInstance(E, EXPR.SumExpressionBase)
        e = E.arg(-1)
        self.assertIs(type(e), EXPR.Numeric_GetItemExpression)
        self.assertIsNot(e, E_base.arg(0))
        self.assertEqual(e.nargs(), 2)
        self.assertIs(e.arg(0), m.x)
        self.assertIsInstance(e.arg(1), EXPR.SumExpressionBase)
        self.assertIs(e.arg(1).arg(0), t)
        self.assertIs(type(e.arg(1).arg(1)), EXPR.NPV_Numeric_GetItemExpression)
        self.assertIs(type(e.arg(1).arg(1)), type(E_base.arg(-1).arg(1).arg(1)))
        self.assertIsNot(e.arg(1).arg(1), E_base.arg(-1).arg(1).arg(1))
        self.assertIsInstance(e.arg(1).arg(1).arg(1), EXPR.SumExpressionBase)
        self.assertIs(e.arg(1).arg(1).arg(1).arg(0), t)


class TestTemplatizeRule(unittest.TestCase):
    def test_simple_rule(self):
        m = ConcreteModel()
        m.I = RangeSet(3)
        m.x = Var(m.I)

        @m.Constraint(m.I)
        def c(m, i):
            return m.x[i] <= 0

        template, indices = templatize_constraint(m.c)
        self.assertEqual(len(indices), 1)
        self.assertIs(indices[0]._set, m.I)
        self.assertEqual(str(template), "x[_1]  <=  0")
        # Test that the RangeSet iterator was put back
        self.assertEqual(list(m.I), list(range(1, 4)))
        # Evaluate the template
        indices[0].set_value(2)
        self.assertEqual(str(resolve_template(template)), 'x[2]  <=  0')

    def test_tuple_rules(self):
        m = ConcreteModel()
        m.I = RangeSet(3)
        m.x = Var(m.I)

        @m.Constraint(m.I)
        def c(m, i):
            return (None, m.x[i], 0)

        template, indices = templatize_constraint(m.c)
        self.assertEqual(len(indices), 1)
        self.assertIs(indices[0]._set, m.I)
        self.assertEqual(str(template), "x[_1]  <=  0")
        # Test that the RangeSet iterator was put back
        self.assertEqual(list(m.I), list(range(1, 4)))
        # Evaluate the template
        indices[0].set_value(2)
        self.assertEqual(str(resolve_template(template)), 'x[2]  <=  0')

        @m.Constraint(m.I)
        def d(m, i):
            return (0, m.x[i], 10)

        template, indices = templatize_constraint(m.d)
        self.assertEqual(len(indices), 1)
        self.assertIs(indices[0]._set, m.I)
        self.assertEqual(str(template), "0  <=  x[_1]  <=  10")
        # Test that the RangeSet iterator was put back
        self.assertEqual(list(m.I), list(range(1, 4)))
        # Evaluate the template
        indices[0].set_value(2)
        self.assertEqual(str(resolve_template(template)), '0  <=  x[2]  <=  10')

        @m.Constraint(m.I)
        def e(m, i):
            return (m.x[i], 0)

        template, indices = templatize_constraint(m.e)
        self.assertEqual(len(indices), 1)
        self.assertIs(indices[0]._set, m.I)
        self.assertEqual(str(template), "x[_1]  ==  0")
        # Test that the RangeSet iterator was put back
        self.assertEqual(list(m.I), list(range(1, 4)))
        # Evaluate the template
        indices[0].set_value(2)
        self.assertEqual(str(resolve_template(template)), 'x[2]  ==  0')

    def test_simple_rule_nonfinite_set(self):
        m = ConcreteModel()
        m.x = Var(Integers, dense=False)

        @m.Constraint(Integers)
        def c(m, i):
            return m.x[i] <= 0

        template, indices = templatize_constraint(m.c)
        self.assertEqual(len(indices), 1)
        self.assertIs(indices[0]._set, Integers)
        self.assertEqual(str(template), "x[_1]  <=  0")
        # Evaluate the template
        indices[0].set_value(2)
        self.assertEqual(str(resolve_template(template)), 'x[2]  <=  0')

    def test_simple_abstract_rule(self):
        m = AbstractModel()
        m.I = RangeSet(3)
        m.x = Var(m.I)

        @m.Constraint(m.I)
        def c(m, i):
            return m.x[i] <= 0

        # Note: the constraint can be abstract, but the Set/Var must
        # have been constructed (otherwise accessing the Set raises an
        # exception)

        with self.assertRaisesRegex(ValueError, ".*has not been constructed"):
            template, indices = templatize_constraint(m.c)

        m.I.construct()
        m.x.construct()
        template, indices = templatize_constraint(m.c)
        self.assertEqual(len(indices), 1)
        self.assertIs(indices[0]._set, m.I)
        self.assertEqual(str(template), "x[_1]  <=  0")

    def test_simple_sum_rule(self):
        m = ConcreteModel()
        m.I = RangeSet(3)
        m.J = RangeSet(3)
        m.x = Var(m.I, m.J)

        @m.Constraint(m.I)
        def c(m, i):
            return sum(m.x[i, j] for j in m.J) <= 0

        template, indices = templatize_constraint(m.c)
        self.assertEqual(len(indices), 1)
        self.assertIs(indices[0]._set, m.I)
        self.assertEqual(
            template.to_string(verbose=True),
            "templatesum(getitem(x, _1, _2), iter(_2, J))  <=  0",
        )
        self.assertEqual(str(template), "SUM(x[_1,_2] for _2 in J)  <=  0")
        # Evaluate the template
        indices[0].set_value(2)
        self.assertEqual(
            str(resolve_template(template)), 'x[2,1] + x[2,2] + x[2,3]  <=  0'
        )

    def test_nested_sum_rule(self):
        m = ConcreteModel()
        m.I = RangeSet(3)
        m.J = RangeSet(3)
        m.K = Set(m.I, initialize={1: [10], 2: [10, 20], 3: [10, 20, 30]})
        m.x = Var(m.I, m.J, [10, 20, 30])

        @m.Constraint()
        def c(m):
            return sum(sum(m.x[i, j, k] for k in m.K[i]) for j in m.J for i in m.I) <= 0

        template, indices = templatize_constraint(m.c)
        self.assertEqual(len(indices), 0)
        self.assertEqual(
            template.to_string(verbose=True),
            "templatesum("
            "templatesum(getitem(x, _2, _1, _3), iter(_3, getitem(K, _2))), "
            "iter(_1, J), iter(_2, I))  <=  0",
        )
        self.assertEqual(
            str(template),
            "SUM(SUM(x[_2,_1,_3] for _3 in K[_2]) for _1 in J for _2 in I)  <=  0",
        )
        # Evaluate the template
        self.assertEqual(
            str(resolve_template(template)),
            'x[1,1,10] + '
            'x[2,1,10] + x[2,1,20] + '
            'x[3,1,10] + x[3,1,20] + x[3,1,30] + '
            'x[1,2,10] + '
            'x[2,2,10] + x[2,2,20] + '
            'x[3,2,10] + x[3,2,20] + x[3,2,30] + '
            'x[1,3,10] + '
            'x[2,3,10] + x[2,3,20] + '
            'x[3,3,10] + x[3,3,20] + x[3,3,30]  <=  0',
        )

    def test_multidim_nested_sum_rule(self):
        m = ConcreteModel()
        m.I = RangeSet(3)
        m.J = RangeSet(3)
        m.JI = m.J * m.I
        m.K = Set(m.I, initialize={1: [10], 2: [10, 20], 3: [10, 20, 30]})
        m.x = Var(m.I, m.J, [10, 20, 30])

        @m.Constraint()
        def c(m):
            return sum(sum(m.x[i, j, k] for k in m.K[i]) for j, i in m.JI) <= 0

        template, indices = templatize_constraint(m.c)
        self.assertEqual(len(indices), 0)
        self.assertEqual(
            template.to_string(verbose=True),
            "templatesum("
            "templatesum(getitem(x, _2, _1, _3), iter(_3, getitem(K, _2))), "
            "iter(_1, _2, JI))  <=  0",
        )
        self.assertEqual(
            str(template),
            "SUM(SUM(x[_2,_1,_3] for _3 in K[_2]) for _1, _2 in JI)  <=  0",
        )
        # Evaluate the template
        self.assertEqual(
            str(resolve_template(template)),
            'x[1,1,10] + '
            '(x[2,1,10] + x[2,1,20]) + '
            '(x[3,1,10] + x[3,1,20] + x[3,1,30]) + '
            '(x[1,2,10]) + '
            '(x[2,2,10] + x[2,2,20]) + '
            '(x[3,2,10] + x[3,2,20] + x[3,2,30]) + '
            '(x[1,3,10]) + '
            '(x[2,3,10] + x[2,3,20]) + '
            '(x[3,3,10] + x[3,3,20] + x[3,3,30])  <=  0',
        )

    def test_multidim_nested_sum_rule(self):
        m = ConcreteModel()
        m.I = RangeSet(3)
        m.J = RangeSet(3)
        m.JI = m.J * m.I
        m.K = Set(m.I, initialize={1: [10], 2: [10, 20], 3: [10, 20, 30]})
        m.x = Var(m.I, m.J, [10, 20, 30])

        @m.Constraint()
        def c(m):
            return sum(sum(m.x[i, j, k] for k in m.K[i]) for j, i in m.JI) <= 0

        template, indices = templatize_constraint(m.c)
        self.assertEqual(len(indices), 0)
        self.assertEqual(
            template.to_string(verbose=True),
            "templatesum("
            "templatesum(getitem(x, _2, _1, _3), iter(_3, getitem(K, _2))), "
            "iter(_1, _2, JI))  <=  0",
        )
        self.assertEqual(
            str(template),
            "SUM(SUM(x[_2,_1,_3] for _3 in K[_2]) for _1, _2 in JI)  <=  0",
        )
        # Evaluate the template
        self.assertEqual(
            str(resolve_template(template)),
            'x[1,1,10] + '
            'x[2,1,10] + x[2,1,20] + '
            'x[3,1,10] + x[3,1,20] + x[3,1,30] + '
            'x[1,2,10] + '
            'x[2,2,10] + x[2,2,20] + '
            'x[3,2,10] + x[3,2,20] + x[3,2,30] + '
            'x[1,3,10] + '
            'x[2,3,10] + x[2,3,20] + '
            'x[3,3,10] + x[3,3,20] + x[3,3,30]  <=  0',
        )

    def test_multidim_nested_getattr_sum_rule(self):
        m = ConcreteModel()
        m.I = RangeSet(3)
        m.J = RangeSet(3)
        m.JI = m.J * m.I
        m.K = Set(m.I, initialize={1: [10], 2: [10, 20], 3: [10, 20, 30]})
        m.x = Var(m.I, m.J, [10, 20, 30])

        @m.Block(m.I)
        def b(b, i):
            b.K = RangeSet(10, 10 * i, 10)

        @m.Constraint()
        def c(m):
            return sum(sum(m.x[i, j, k] for k in m.b[i].K) for j, i in m.JI) <= 0

        template, indices = templatize_constraint(m.c)
        self.assertEqual(len(indices), 0)
        self.assertEqual(
            template.to_string(verbose=True),
            "templatesum("
            "templatesum(getitem(x, _2, _1, _3), "
            "iter(_3, getattr(getitem(b, _2), 'K'))), "
            "iter(_1, _2, JI))  <=  0",
        )
        self.assertEqual(
            str(template),
            "SUM(SUM(x[_2,_1,_3] for _3 in b[_2].K) for _1, _2 in JI)  <=  0",
        )
        # Evaluate the template
        self.assertEqual(
            str(resolve_template(template)),
            'x[1,1,10] + '
            'x[2,1,10] + x[2,1,20] + '
            'x[3,1,10] + x[3,1,20] + x[3,1,30] + '
            'x[1,2,10] + '
            'x[2,2,10] + x[2,2,20] + '
            'x[3,2,10] + x[3,2,20] + x[3,2,30] + '
            'x[1,3,10] + '
            'x[2,3,10] + x[2,3,20] + '
            'x[3,3,10] + x[3,3,20] + x[3,3,30]  <=  0',
        )

    def test_eval_getattr(self):
        m = ConcreteModel()
        m.T = RangeSet(3)

        @m.Block(m.T)
        def b(b, i):
            b.x = Var(initialize=i)

            @b.Block(m.T)
            def bb(bb, j):
                bb.I = RangeSet(i * j)
                bb.y = Var(bb.I, initialize=lambda m, i: i)

        t = IndexTemplate(m.T)
        e = m.b[t].x

        with self.assertRaisesRegex(
            ValueError, r'Evaluating uninitialized IndexTemplate \({T}\)'
        ):
            value(e())
        with self.assertRaisesRegex(
            KeyError, r"Index 'None' is not valid for indexed component 'b'"
        ):
            self.assertIsNone(e(exception=False))
        with self.assertRaisesRegex(
            KeyError, r"Index 'None' is not valid for indexed component 'b'"
        ):
            self.assertIsNone(e(False))

        t.set_value(2)
        self.assertEqual(e(), 2)

        f = e.set_value(5)
        self.assertIs(f.__class__, CallExpression)
        self.assertEqual(f._kwds, ())
        self.assertEqual(len(f._args_), 2)
        self.assertIs(f._args_[0].__class__, EXPR.Structural_GetAttrExpression)
        self.assertIs(f._args_[0]._args_[0], e)
        self.assertEqual(f._args_[1], 5)

        self.assertEqual(value(m.b[2].x), 2)
        f()
        self.assertEqual(value(m.b[2].x), 5)

        f = e.set_value('a', skip_validation=True)
        self.assertIs(f.__class__, CallExpression)
        self.assertEqual(f._kwds, ('skip_validation',))
        self.assertEqual(len(f._args_), 3)
        self.assertIs(f._args_[0].__class__, EXPR.Structural_GetAttrExpression)
        self.assertIs(f._args_[0]._args_[0], e)
        self.assertEqual(f._args_[1], 'a')
        self.assertEqual(f._args_[2], True)

        f()
        self.assertEqual(value(m.b[2].x), 'a')


class TestTemplateSubstitution(unittest.TestCase):
    def setUp(self):
        self.m = m = ConcreteModel()
        m.TRAY = Set(initialize=range(5))
        m.TIME = Set(bounds=(0, 10), initialize=range(10))
        m.y = Var(initialize=1)
        m.x = Var(m.TIME, m.TRAY, initialize=lambda _m, i, j: i)
        m.dxdt = Var(m.TIME, m.TRAY, initialize=lambda _m, i, j: 2 * i)

    def test_simple_substitute_param(self):
        def diffeq(m, t, i):
            return (
                m.dxdt[t, i]
                == t * m.x[t, i - 1] ** 2 + m.y**2 + m.x[t, i + 1] + m.x[t, i - 1]
            )

        m = self.m
        t = IndexTemplate(m.TIME)
        e = diffeq(m, t, 2)

        self.assertTrue(isinstance(e, EXPR.RelationalExpression))

        _map = {}
        E = substitute_template_expression(e, substitute_getitem_with_param, _map)
        self.assertIsNot(e, E)

        self.assertEqual(len(_map), 3)

        idx1 = _GetItemIndexer(m.x[t, 1])
        self.assertEqual(idx1.nargs(), 2)
        self.assertIs(idx1.base, m.x)
        self.assertIs(idx1.arg(0), t)
        self.assertEqual(idx1.arg(1), 1)
        self.assertIn(idx1, _map)

        idx2 = _GetItemIndexer(m.dxdt[t, 2])
        self.assertEqual(idx2.nargs(), 2)
        self.assertIs(idx2.base, m.dxdt)
        self.assertIs(idx2.arg(0), t)
        self.assertEqual(idx2.arg(1), 2)
        self.assertIn(idx2, _map)

        idx3 = _GetItemIndexer(m.x[t, 3])
        self.assertEqual(idx3.nargs(), 2)
        self.assertIs(idx3.base, m.x)
        self.assertIs(idx3.arg(0), t)
        self.assertEqual(idx3.arg(1), 3)
        self.assertIn(idx3, _map)

        self.assertFalse(idx1 == idx2)
        self.assertFalse(idx1 == idx3)
        self.assertFalse(idx2 == idx3)

        idx4 = _GetItemIndexer(m.x[t, 2])
        self.assertNotIn(idx4, _map)

        t.set_value(5)
        self.assertEqual((e.arg(0)(), e.arg(1)()), (10, 136))

        self.assertEqual(
            str(E),
            "'dxdt[{TIME},2]'  ==  "
            "{TIME}*'x[{TIME},1]'**2 + y**2 + 'x[{TIME},3]' + 'x[{TIME},1]'",
        )

        _map[idx1].set_value(value(m.x[value(t), 1]))
        _map[idx2].set_value(value(m.dxdt[value(t), 2]))
        _map[idx3].set_value(value(m.x[value(t), 3]))
        self.assertEqual((E.arg(0)(), E.arg(1)()), (10, 136))

        _map[idx1].set_value(12)
        _map[idx2].set_value(34)
        self.assertEqual((E.arg(0)(), E.arg(1)()), (34, 738))

    def test_simple_substitute_index(self):
        def diffeq(m, t, i):
            return m.dxdt[t, i] == t * m.x[t, i] ** 2 + m.y**2

        m = self.m
        t = IndexTemplate(m.TIME)
        e = diffeq(m, t, 2)
        t.set_value(5)

        self.assertTrue(isinstance(e, EXPR.RelationalExpression))
        self.assertEqual((e.arg(0)(), e.arg(1)()), (10, 126))

        E = substitute_template_expression(e, substitute_template_with_value)
        self.assertIsNot(e, E)

        self.assertEqual(str(E), 'dxdt[5,2]  ==  5.0*x[5,2]**2 + y**2')


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.log import LoggingIntercept
import pyomo.common.unittest as unittest

from pyomo.core.expr.compare import assertExpressionsEqual
from pyomo.core.expr.numeric_expr import (
    LinearExpression,
    MonomialTermExpression,
    SumExpression,
)
from pyomo.repn.quadratic import QuadraticRepnVisitor
from pyomo.repn.tests.test_linear import VisitorConfig
from pyomo.repn.util import InvalidNumber

from pyomo.environ import ConcreteModel, Var, Param, Any, log


class TestQuadratic(unittest.TestCase):
    def test_product(self):
        m = ConcreteModel()
        m.x = Var()
        m.y = Var()

        e = 2

        cfg = VisitorConfig()
        visitor = QuadraticRepnVisitor(**cfg)
        visitor.expand_nonlinear_products = True
        repn = visitor.walk_expression(e)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 2)
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.quadratic, None)
        self.assertEqual(repn.nonlinear, None)

        e = 2 + 3 * m.x

        cfg = VisitorConfig()
        visitor = QuadraticRepnVisitor(**cfg)
        visitor.expand_nonlinear_products = True
        repn = visitor.walk_expression(e)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 2)
        self.assertEqual(repn.linear, {id(m.x): 3})
        self.assertEqual(repn.quadratic, None)
        self.assertEqual(repn.nonlinear, None)

        e = 2 + 3 * m.x + 4 * m.x**2

        cfg = VisitorConfig()
        visitor = QuadraticRepnVisitor(**cfg)
        visitor.expand_nonlinear_products = True
        repn = visitor.walk_expression(e)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 2)
        self.assertEqual(repn.linear, {id(m.x): 3})
        self.assertEqual(repn.quadratic, {(id(m.x), id(m.x)): 4})
        self.assertEqual(repn.nonlinear, None)

        e = (2 + 3 * m.x + 4 * m.x**2) * (5 + 6 * m.x + 7 * m.x**2)

        cfg = VisitorConfig()
        visitor = QuadraticRepnVisitor(**cfg)
        visitor.expand_nonlinear_products = True
        repn = visitor.walk_expression(e)

        QE4 = 4 * m.x**2
        QE7 = 7 * m.x**2
        LE3 = MonomialTermExpression((3, m.x))
        LE6 = MonomialTermExpression((6, m.x))
        NL = +QE4 * (QE7 + LE6) + (LE3) * (QE7)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 10)
        self.assertEqual(repn.linear, {id(m.x): 27})
        self.assertEqual(repn.quadratic, {(id(m.x), id(m.x)): 52})
        assertExpressionsEqual(self, repn.nonlinear, NL)

        e = (2 + 3 * m.x + 4 * m.x**2) * (5 + 6 * m.x + 7 * m.x**2)

        cfg = VisitorConfig()
        visitor = QuadraticRepnVisitor(**cfg)
        visitor.expand_nonlinear_products = False
        repn = visitor.walk_expression(e)

        NL = (4 * m.x**2 + 3 * m.x + 2) * (7 * m.x**2 + 6 * m.x + 5)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.quadratic, None)
        assertExpressionsEqual(self, repn.nonlinear, NL)

        e = (1 + 2 * m.x + 3 * m.y) * (4 + 5 * m.x + 6 * m.y)

        cfg = VisitorConfig()
        repn = QuadraticRepnVisitor(**cfg).walk_expression(e)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x, id(m.y): m.y})
        self.assertEqual(cfg.var_order, {id(m.x): 0, id(m.y): 1})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 4)
        self.assertEqual(repn.linear, {id(m.x): 13, id(m.y): 18})
        self.assertEqual(
            cfg.order_quadratic(repn.quadratic),
            {(id(m.x), id(m.x)): 10, (id(m.y), id(m.y)): 18, (id(m.x), id(m.y)): 27},
        )
        assertExpressionsEqual(self, repn.nonlinear, None)

        e = (m.x + m.y + log(m.x)) * m.x

        cfg = VisitorConfig()
        visitor = QuadraticRepnVisitor(**cfg)
        visitor.expand_nonlinear_products = False
        repn = visitor.walk_expression(e)

        NL = (log(m.x) + (m.x + m.y)) * m.x

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x, id(m.y): m.y})
        self.assertEqual(cfg.var_order, {id(m.x): 0, id(m.y): 1})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.quadratic, None)
        assertExpressionsEqual(self, repn.nonlinear, NL)

        visitor.expand_nonlinear_products = True
        repn = visitor.walk_expression(e)

        NL = log(m.x) * m.x

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x, id(m.y): m.y})
        self.assertEqual(cfg.var_order, {id(m.x): 0, id(m.y): 1})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {})
        self.assertEqual(
            cfg.order_quadratic(repn.quadratic),
            {(id(m.x), id(m.x)): 1, (id(m.x), id(m.y)): 1},
        )
        assertExpressionsEqual(self, repn.nonlinear, NL)

        e = m.x * (m.x + m.y + log(m.x) + 2)

        cfg = VisitorConfig()
        visitor = QuadraticRepnVisitor(**cfg)
        visitor.expand_nonlinear_products = False
        repn = visitor.walk_expression(e)

        NL = m.x * (log(m.x) + (m.x + m.y + 2))

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x, id(m.y): m.y})
        self.assertEqual(cfg.var_order, {id(m.x): 0, id(m.y): 1})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.quadratic, None)
        assertExpressionsEqual(self, repn.nonlinear, NL)

        visitor.expand_nonlinear_products = True
        repn = visitor.walk_expression(e)

        NL = m.x * log(m.x)

        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x, id(m.y): m.y})
        self.assertEqual(cfg.var_order, {id(m.x): 0, id(m.y): 1})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {id(m.x): 2})
        self.assertEqual(
            cfg.order_quadratic(repn.quadratic),
            {(id(m.x), id(m.x)): 1, (id(m.x), id(m.y)): 1},
        )
        assertExpressionsEqual(self, repn.nonlinear, NL)

    def test_sum(self):
        m = ConcreteModel()
        m.x = Var()
        m.y = Var()

        e = SumExpression([])

        cfg = VisitorConfig()
        repn = QuadraticRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.quadratic, None)
        self.assertEqual(repn.nonlinear, None)

        e += 5

        cfg = VisitorConfig()
        repn = QuadraticRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {})
        self.assertEqual(cfg.var_order, {})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 5)
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.quadratic, None)
        self.assertEqual(repn.nonlinear, None)

        e += m.x

        cfg = VisitorConfig()
        repn = QuadraticRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x})
        self.assertEqual(cfg.var_order, {id(m.x): 0})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 5)
        self.assertEqual(repn.linear, {id(m.x): 1})
        self.assertEqual(repn.quadratic, None)
        self.assertEqual(repn.nonlinear, None)

        e += m.y**2

        cfg = VisitorConfig()
        repn = QuadraticRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x, id(m.y): m.y})
        self.assertEqual(cfg.var_order, {id(m.x): 0, id(m.y): 1})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 5)
        self.assertEqual(repn.linear, {id(m.x): 1})
        self.assertEqual(repn.quadratic, {(id(m.y), id(m.y)): 1})
        self.assertEqual(repn.nonlinear, None)

        e += m.y**3

        cfg = VisitorConfig()
        repn = QuadraticRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x, id(m.y): m.y})
        self.assertEqual(cfg.var_order, {id(m.x): 0, id(m.y): 1})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 5)
        self.assertEqual(repn.linear, {id(m.x): 1})
        self.assertEqual(repn.quadratic, {(id(m.y), id(m.y)): 1})
        assertExpressionsEqual(self, repn.nonlinear, m.y**3)

        e += 2 * m.x**4

        cfg = VisitorConfig()
        repn = QuadraticRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x, id(m.y): m.y})
        self.assertEqual(cfg.var_order, {id(m.x): 0, id(m.y): 1})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 5)
        self.assertEqual(repn.linear, {id(m.x): 1})
        self.assertEqual(repn.quadratic, {(id(m.y), id(m.y)): 1})
        assertExpressionsEqual(self, repn.nonlinear, m.y**3 + 2 * m.x**4)

        e += 2 * m.y

        cfg = VisitorConfig()
        repn = QuadraticRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x, id(m.y): m.y})
        self.assertEqual(cfg.var_order, {id(m.x): 0, id(m.y): 1})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 5)
        self.assertEqual(repn.linear, {id(m.x): 1, id(m.y): 2})
        self.assertEqual(repn.quadratic, {(id(m.y), id(m.y)): 1})
        assertExpressionsEqual(self, repn.nonlinear, m.y**3 + 2 * m.x**4)

        e += 3 * m.x * m.y

        cfg = VisitorConfig()
        repn = QuadraticRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x, id(m.y): m.y})
        self.assertEqual(cfg.var_order, {id(m.x): 0, id(m.y): 1})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 5)
        self.assertEqual(repn.linear, {id(m.x): 1, id(m.y): 2})
        self.assertEqual(
            cfg.order_quadratic(repn.quadratic),
            {(id(m.y), id(m.y)): 1, (id(m.x), id(m.y)): 3},
        )
        assertExpressionsEqual(self, repn.nonlinear, m.y**3 + 2 * m.x**4)

    def test_pow(self):
        m = ConcreteModel()
        m.x = Var()
        m.y = Var()

        # Check **{int}
        cfg = VisitorConfig()
        repn = QuadraticRepnVisitor(**cfg).walk_expression((1 + 3 * m.x + 4 * m.y) ** 2)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x, id(m.y): m.y})
        self.assertEqual(cfg.var_order, {id(m.x): 0, id(m.y): 1})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 1)
        self.assertEqual(repn.linear, {id(m.x): 6, id(m.y): 8})
        self.assertEqual(
            cfg.order_quadratic(repn.quadratic),
            {(id(m.x), id(m.x)): 9, (id(m.y), id(m.y)): 16, (id(m.x), id(m.y)): 24},
        )
        self.assertEqual(repn.nonlinear, None)

        # Check **{int}
        cfg = VisitorConfig()
        repn = QuadraticRepnVisitor(**cfg).walk_expression(
            (1 + 3 * m.x + 4 * m.y) ** 2.0
        )
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(cfg.var_map, {id(m.x): m.x, id(m.y): m.y})
        self.assertEqual(cfg.var_order, {id(m.x): 0, id(m.y): 1})
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 1)
        self.assertEqual(repn.linear, {id(m.x): 6, id(m.y): 8})
        self.assertEqual(
            cfg.order_quadratic(repn.quadratic),
            {(id(m.x), id(m.x)): 9, (id(m.y), id(m.y)): 16, (id(m.x), id(m.y)): 24},
        )
        self.assertEqual(repn.nonlinear, None)

    def test_zero_elimination(self):
        m = ConcreteModel()
        m.x = Var(range(4))

        e = 0 * m.x[0] + 0 * m.x[1] * m.x[2] + 0 * log(m.x[3])

        cfg = VisitorConfig()
        repn = QuadraticRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(
            cfg.var_map,
            {
                id(m.x[0]): m.x[0],
                id(m.x[1]): m.x[1],
                id(m.x[2]): m.x[2],
                id(m.x[3]): m.x[3],
            },
        )
        self.assertEqual(
            cfg.var_order, {id(m.x[0]): 0, id(m.x[1]): 1, id(m.x[2]): 2, id(m.x[3]): 3}
        )
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {})
        self.assertEqual(repn.quadratic, None)
        assertExpressionsEqual(self, repn.nonlinear, 0 * log(m.x[3]))

        m.p = Param(mutable=True, within=Any, initialize=None)
        e = m.p * m.x[0] + m.p * m.x[1] * m.x[2] + m.p * log(m.x[3])

        cfg = VisitorConfig()
        repn = QuadraticRepnVisitor(**cfg).walk_expression(e)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(
            cfg.var_map,
            {
                id(m.x[0]): m.x[0],
                id(m.x[1]): m.x[1],
                id(m.x[2]): m.x[2],
                id(m.x[3]): m.x[3],
            },
        )
        self.assertEqual(
            cfg.var_order, {id(m.x[0]): 0, id(m.x[1]): 1, id(m.x[2]): 2, id(m.x[3]): 3}
        )
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 0)
        self.assertEqual(repn.linear, {id(m.x[0]): InvalidNumber(None)})
        self.assertEqual(
            cfg.order_quadratic(repn.quadratic),
            {(id(m.x[1]), id(m.x[2])): InvalidNumber(None)},
        )
        self.assertEqual(repn.nonlinear, InvalidNumber(None))

        e = (
            m.p * m.x[0]
            + m.p * m.x[1]
            + m.p * m.x[1] * m.x[2]
            + m.p * m.x[2] ** 2
            + m.p * log(m.x[3])
        )
        f = 1 + m.x[0] + m.x[2] ** 2 + 0 * e

        cfg = VisitorConfig()
        repn = QuadraticRepnVisitor(**cfg).walk_expression(f)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(
            cfg.var_map,
            {
                id(m.x[0]): m.x[0],
                id(m.x[1]): m.x[1],
                id(m.x[2]): m.x[2],
                id(m.x[3]): m.x[3],
            },
        )
        self.assertEqual(
            cfg.var_order, {id(m.x[0]): 0, id(m.x[1]): 1, id(m.x[2]): 2, id(m.x[3]): 3}
        )
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, 1)
        self.assertEqual(len(repn.linear), 2)
        self.assertEqual(
            repn.linear,
            {id(m.x[0]): InvalidNumber(None), id(m.x[1]): InvalidNumber(None)},
        )
        self.assertEqual(len(repn.quadratic), 2)
        self.assertEqual(
            cfg.order_quadratic(repn.quadratic),
            {
                (id(m.x[1]), id(m.x[2])): InvalidNumber(None),
                (id(m.x[2]), id(m.x[2])): InvalidNumber(None),
            },
        )
        self.assertEqual(repn.nonlinear, InvalidNumber(None))

        f = 1 + m.p + 0 * e

        cfg = VisitorConfig()
        repn = QuadraticRepnVisitor(**cfg).walk_expression(f)
        self.assertEqual(cfg.subexpr, {})
        self.assertEqual(
            cfg.var_map,
            {
                id(m.x[0]): m.x[0],
                id(m.x[1]): m.x[1],
                id(m.x[2]): m.x[2],
                id(m.x[3]): m.x[3],
            },
        )
        self.assertEqual(
            cfg.var_order, {id(m.x[0]): 0, id(m.x[1]): 1, id(m.x[2]): 2, id(m.x[3]): 3}
        )
        self.assertEqual(repn.multiplier, 1)
        self.assertEqual(repn.constant, InvalidNumber(None))
        self.assertEqual(len(repn.linear), 2)
        self.assertEqual(
            repn.linear,
            {id(m.x[0]): InvalidNumber(None), id(m.x[1]): InvalidNumber(None)},
        )
        self.assertEqual(len(repn.quadratic), 2)
        self.assertEqual(
            cfg.order_quadratic(repn.quadratic),
            {
                (id(m.x[1]), id(m.x[2])): InvalidNumber(None),
                (id(m.x[2]), id(m.x[2])): InvalidNumber(None),
            },
        )
        self.assertEqual(repn.nonlinear, InvalidNumber(None))



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.dependencies import numpy as np
from pyomo.common.dependencies import scipy, attempt_import

import pyomo.environ as pyo

# Use attempt_import here due to unguarded NumPy import in this file
nlp = attempt_import('pyomo.contrib.pynumero.interfaces.pyomo_nlp')[0]
from pyomo.common.collections import ComponentSet, ComponentMap
import pyomo.core.expr.calculus.derivatives as derivatives


def _coo_reorder_cols(mat, remap):
    """Change the order of columns in a COO matrix. The main use of this is
    to reorder variables in the Jacobian matrix. This changes the matrix in
    place.

    Parameters
    ----------
    mat: scipy.sparse.coo_matrix
        Reorder the columns of this matrix
    remap: dict
        dictionary where keys are old column and value is new column, if a column
        doesn't move, it doesn't need to be included.

    Returns
    -------
    NoneType
        None
    """
    for i in range(len(mat.data)):
        try:
            mat.col[i] = remap[mat.col[i]]
        except KeyError:
            pass  # it's fine if we don't move a col in remap


def get_dsdp_dfdp(model, theta):
    """Calculate the derivatives of the state variables (s) with respect to
    parameters (p) (ds/dp), and the derivative of the objective function (f)
    with respect to p (df/dp). The number of parameters in theta should be the
    same as the number of degrees of freedom.

    Parameters
    ----------
    model: pyomo.environ.Block | pyomo.contrib.pynumero.interfaces.PyomoNLP
        Model to calculate sensitivity on. To retain the cached objects in
        the pynumero interface, create a PyomoNLP first and pass it to this function.
    theta: list
        A list of parameters as pyomo.environ.VarData, the number of parameters
        should be equal to the degrees of freedom.

    Returns
    -------
    scipy.sparse.csc_matrix, csc_matrix, ComponentMap, ComponentMap
        ds/dp (ns + np by np), df/dp (1 by np), row map, column map.
        The column map maps Pyomo variables p to columns and the
        row map maps Pyomo variables s to rows.
    """
    # Create a Pynumero NLP and get Jacobian
    if isinstance(model, nlp.PyomoNLP):
        m2 = model
    else:
        m2 = nlp.PyomoNLP(model)
    J = m2.evaluate_jacobian_eq()
    v_list = m2.get_pyomo_variables()
    # Map variables to columns in J
    mv_map = {id(v): i for i, v in enumerate(v_list)}
    s_list = list(ComponentSet(v_list) - ComponentSet(theta))
    ns = len(s_list)
    np = len(theta)
    col_remap = {mv_map[id(v)]: i for i, v in enumerate(s_list + theta)}
    _coo_reorder_cols(J, remap=col_remap)
    J = J.tocsc()
    dB = -(
        J
        @ scipy.sparse.vstack(
            (scipy.sparse.coo_matrix((ns, np)), scipy.sparse.identity(np))
        ).tocsc()
    )
    # Calculate sensitivity matrix
    dsdp = scipy.sparse.linalg.spsolve(J[:, range(ns)], dB)
    # Get a map of state vars to columns
    s_map = {id(v): i for i, v in enumerate(s_list)}
    # Get the outputs we are interested in from the list of output vars
    column_map = ComponentMap([(v, i) for i, v in enumerate(theta)])
    row_map = ComponentMap([(v, i) for i, v in enumerate(s_list)])
    dfdx = scipy.sparse.coo_matrix(m2.evaluate_grad_objective())
    _coo_reorder_cols(dfdx, remap=col_remap)
    dfdx = dfdx.tocsc()
    dfdp = dfdx[0, :ns] @ dsdp + dfdx[0, ns:]

    # Add independent variables to the rows as the end
    dsdp = scipy.sparse.vstack(
        [dsdp, scipy.sparse.identity(len(column_map), format="csc")]
    )
    n = len(row_map)
    for p, i in column_map.items():
        row_map[p] = i + n

    # return sensitivity of the outputs to p and component maps
    return dsdp, dfdp, row_map, column_map


def get_dydp(y_list, dsdp, row_map, column_map=None):
    """Reduce the sensitivity matrix from get_dsdp_dfdp to only
    a specified set of state variables of interest.

    Parameters
    ----------
    y_list: list
        A list of state variables or named expressions
    dsdp: csc_matrix
        A sensitivity matrix calculated by get_dsdp_dfdp
    row_map: ComponentMap
        A row map from get_dsdp_dfdp
    column_map: ComponentMap
        A column map from get_dsdp_dfdp, only needed if y_list
        contains expressions

    Returns
    -------
    csc_matrix, ComponentMap
        dy/dp and a new row map with only y variables

    """
    j = 0
    new_row_map = ComponentMap()
    expr_row_map = ComponentMap()
    rows = [None] * len(y_list)
    for i, v in enumerate(y_list):
        new_row_map[v] = i
        if v not in row_map:
            expr_row_map[v] = j
            j += 1
        else:
            rows[i] = dsdp[row_map[v], :]

    if j > 0:
        if column_map is None:
            raise ValueError(
                "A column_map must be provided if y_list contains named expressions."
            )
        wrt_list = [s for s in row_map]
        dedx = [None] * j
        for v, i in expr_row_map.items():
            dedx[i] = scipy.sparse.csc_matrix(
                derivatives.differentiate(v, wrt_list=wrt_list)
            )
        dedx = scipy.sparse.vstack(dedx)
        ns = len(row_map) - len(column_map)
        dedp = dedx[:, :ns] @ dsdp[:ns, :] + dedx[:, ns:]
        for v, i in expr_row_map.items():
            rows[new_row_map[v]] = dedp[i, :]

    dydp = scipy.sparse.vstack(rows)
    return dydp, new_row_map
